from typing import Callable, List, Tuple

import numpy as np
from numba import cuda
from numpy.testing import assert_allclose
from pytest import approx, mark
from scipy.spatial.transform import Rotation

from sarracen import SarracenDataFrame
from sarracen.kernels import CubicSplineKernel
from sarracen.interpolate import interpolate_3d_cross, interpolate_3d_proj, \
    interpolate_3d_vec, interpolate_3d_cross_vec, interpolate_3d_grid

backends = ['cpu']
if cuda.is_available():
    backends.append('gpu')


def rotate(target: Tuple[float, float, float],
           rot_z: float,
           rot_y: float,
           rot_x: float) -> Tuple[float, float, float]:
    """ Perform a rotation of a target vector in three dimensions.

    A helper function for test_nonstandard_rotation()

    Parameters
    ----------
    target: float tuple of shape (3)
    rot_z, rot_y, rot_x: Rotation around each axis (in degrees)

    Returns
    -------
    float tuple of shape (3):
        The rotated vector.
    """
    r = 180 / np.pi

    pos_x1 = target[0] * np.cos(rot_z / r) - target[1] * np.sin(rot_z / r)
    pos_y1 = target[0] * np.sin(rot_z / r) + target[1] * np.cos(rot_z / r)
    pos_z1 = target[2]

    pos_x2 = pos_x1 * np.cos(rot_y / r) + pos_z1 * np.sin(rot_y / r)
    pos_y2 = pos_y1
    pos_z2 = pos_x1 * -np.sin(rot_y / r) + pos_z1 * np.cos(rot_y / r)

    pos_x3 = pos_x2
    pos_y3 = pos_y2 * np.cos(rot_x / r) - pos_z2 * np.sin(rot_x / r)
    pos_z3 = pos_y2 * np.sin(rot_x / r) + pos_z2 * np.cos(rot_x / r)

    return pos_x3, pos_y3, pos_z3


@mark.parametrize("backend", backends)
def test_nonstandard_rotation(backend: str) -> None:
    """
    Interpolation of a rotated dataframe with nonstandard angles.
    """
    data = {'x': [1], 'y': [1], 'z': [1],
            'A': [4], 'B': [5], 'C': [6],
            'h': [0.9], 'rho': [0.4], 'm': [0.03]}
    sdf = SarracenDataFrame(data, params=dict())
    kernel = CubicSplineKernel()
    sdf.kernel = kernel
    sdf.backend = backend

    column_kernel = kernel.get_column_kernel_func(1000)

    rot_z, rot_y, rot_x = 129, 34, 50

    img_col = interpolate_3d_proj(sdf, 'A',
                                  x_pixels=50, y_pixels=50,
                                  xlim=(-1, 1), ylim=(-1, 1),
                                  rotation=[rot_z, rot_y, rot_x],
                                  rot_origin=[0, 0, 0],
                                  dens_weight=False,
                                  normalize=False, hmin=False)
    img_xsec = interpolate_3d_cross(sdf, 'A',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    z_slice=0,
                                    rotation=[rot_z, rot_y, rot_x],
                                    rot_origin=[0, 0, 0],
                                    normalize=False, hmin=False)
    img_colvec = interpolate_3d_vec(sdf, 'A', 'B', 'C',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    rotation=[rot_z, rot_y, rot_x],
                                    rot_origin=[0, 0, 0],
                                    dens_weight=False,
                                    normalize=False, hmin=False)
    img_xsecvec = interpolate_3d_cross_vec(sdf, 'A', 'B', 'C',
                                           x_pixels=50, y_pixels=50,
                                           xlim=(-1, 1), ylim=(-1, 1),
                                           z_slice=0,
                                           rotation=[rot_z, rot_y, rot_x],
                                           rot_origin=[0, 0, 0],
                                           normalize=False, hmin=False)

    w_col = sdf['m'] / (sdf['rho'] * sdf['h'] ** 2)
    w_xsec = sdf['m'] / (sdf['rho'] * sdf['h'] ** 3)

    pos_x, pos_y, pos_z = rotate((1, 1, 1), rot_z, rot_y, rot_x)
    target_x, target_y, target_z = rotate((4, 5, 6), rot_z, rot_y, rot_x)

    real = -1 + (np.arange(0, 50) + 0.5) * (1 / 25)

    for y in range(50):
        for x in range(50):
            r = np.sqrt((pos_x - real[x]) ** 2
                        + (pos_y - real[y]) ** 2)
            w = column_kernel(r / sdf['h'][0], 3)
            assert img_col[y][x] == approx(w_col[0] * sdf['A'][0] * w)
            assert img_colvec[0][y][x] == approx(w_col[0] * target_x * w)
            assert img_colvec[1][y][x] == approx(w_col[0] * target_y * w)

            r = np.sqrt((pos_x - real[x])**2
                        + (pos_y - real[y])**2
                        + pos_z**2)
            w = kernel.w(r / sdf['h'][0], 3)
            assert img_xsec[y][x] == approx(w_xsec[0] * sdf['A'][0] * w)
            assert img_xsecvec[0][y][x] == approx(w_xsec[0] * target_x * w)
            assert img_xsecvec[1][y][x] == approx(w_xsec[0] * target_y * w)

    img_grid = interpolate_3d_grid(sdf, 'A',
                                   x_pixels=50, y_pixels=50, z_pixels=50,
                                   xlim=(-1, 1), ylim=(-1, 1), zlim=(-1, 1),
                                   rotation=[rot_z, rot_y, rot_x],
                                   rot_origin=[0, 0, 0],
                                   normalize=False, hmin=False)

    for z in range(50):
        for y in range(50):
            for x in range(50):
                r = np.sqrt((pos_x - real[x])**2
                            + (pos_y - real[y])**2
                            + (pos_z - real[z])**2)
                w = kernel.w(r / sdf['h'][0], 3)
                grid_value = w_xsec[0] * sdf['A'][0] * w

                assert img_grid[z][y][x] == approx(grid_value)


@mark.parametrize("backend", backends)
def test_scipy_rotation_equivalency(backend: str) -> None:
    """
    For interpolation functions, a [z, y, x] rotation defined with degrees
    should be equivalent to the scipy version using from_euler().
    """
    data = {'x': [1], 'y': [1], 'z': [1],
            'A': [4], 'B': [3], 'C': [2],
            'h': [0.9], 'rho': [0.4], 'm': [0.03]}
    sdf = SarracenDataFrame(data, params=dict())
    kernel = CubicSplineKernel()
    sdf.kernel = kernel
    sdf.backend = backend

    rotation = [67, -34, 91]
    scipy_rotation = Rotation.from_euler('zyx', rotation, degrees=True)

    img1 = interpolate_3d_proj(sdf, 'A',
                               x_pixels=50, y_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1),
                               rotation=rotation,
                               rot_origin=[0, 0, 0],
                               normalize=False, hmin=False)
    img2 = interpolate_3d_proj(sdf, 'A',
                               x_pixels=50, y_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1),
                               rotation=scipy_rotation,
                               rot_origin=[0, 0, 0],
                               normalize=False, hmin=False)
    assert_allclose(img1, img2)

    img1 = interpolate_3d_cross(sdf, 'A',
                                x_pixels=50, y_pixels=50,
                                xlim=(-1, 1), ylim=(-1, 1),
                                z_slice=0,
                                rotation=rotation,
                                rot_origin=[0, 0, 0],
                                normalize=False, hmin=False)
    img2 = interpolate_3d_cross(sdf, 'A',
                                x_pixels=50, y_pixels=50,
                                xlim=(-1, 1), ylim=(-1, 1),
                                z_slice=0,
                                rotation=scipy_rotation,
                                rot_origin=[0, 0, 0],
                                normalize=False, hmin=False)
    assert_allclose(img1, img2)

    img1_tuple = interpolate_3d_vec(sdf, 'A', 'B', 'C',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    rotation=rotation,
                                    rot_origin=[0, 0, 0],
                                    normalize=False, hmin=False)
    img2_tuple = interpolate_3d_vec(sdf, 'A', 'B', 'C',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    rotation=scipy_rotation,
                                    rot_origin=[0, 0, 0],
                                    normalize=False,
                                    hmin=False)
    assert_allclose(img1_tuple[0], img2_tuple[0])
    assert_allclose(img1_tuple[1], img2_tuple[1])

    img1_tuple = interpolate_3d_cross_vec(sdf, 'A', 'B', 'C',
                                          x_pixels=50, y_pixels=50,
                                          xlim=(-1, 1), ylim=(-1, 1),
                                          z_slice=0,
                                          rotation=rotation,
                                          rot_origin=[0, 0, 0],
                                          normalize=False, hmin=False)
    img2_tuple = interpolate_3d_cross_vec(sdf, 'A', 'B', 'C',
                                          x_pixels=50, y_pixels=50,
                                          xlim=(-1, 1), ylim=(-1, 1),
                                          z_slice=0,
                                          rotation=scipy_rotation,
                                          rot_origin=[0, 0, 0],
                                          normalize=False, hmin=False)
    assert_allclose(img1_tuple[0], img2_tuple[0])
    assert_allclose(img1_tuple[1], img2_tuple[1])

    img1 = interpolate_3d_grid(sdf, 'A',
                               x_pixels=50, y_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1), zlim=(-1, 1),
                               rotation=rotation,
                               rot_origin=[0, 0, 0],
                               normalize=False, hmin=False)
    img2 = interpolate_3d_grid(sdf, 'A',
                               x_pixels=50, y_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1), zlim=(-1, 1),
                               rotation=scipy_rotation,
                               rot_origin=[0, 0, 0],
                               normalize=False, hmin=False)
    assert_allclose(img1, img2)


@mark.parametrize("backend", backends)
def test_quaternion_rotation(backend: str) -> None:
    """
    Test quaternion rotation defined using scipy.
    """
    data = {'x': [1], 'y': [1], 'z': [1],
            'A': [4], 'B': [3], 'C': [2],
            'h': [1.9], 'rho': [0.4], 'm': [0.03]}
    sdf = SarracenDataFrame(data, params=dict())
    kernel = CubicSplineKernel()
    sdf.kernel = kernel
    sdf.backend = backend

    column_kernel = kernel.get_column_kernel_func(1000)

    quat = Rotation.from_quat([5, 3, 8, 1])
    img_col = interpolate_3d_proj(sdf, 'A',
                                  x_pixels=50, y_pixels=50,
                                  xlim=(-1, 1), ylim=(-1, 1),
                                  rotation=quat,
                                  rot_origin=[0, 0, 0],
                                  dens_weight=False,
                                  normalize=False, hmin=False)
    img_colvec = interpolate_3d_vec(sdf, 'A', 'B', 'C',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    rotation=quat,
                                    rot_origin=[0, 0, 0],
                                    dens_weight=False,
                                    normalize=False, hmin=False)
    img_xsec = interpolate_3d_cross(sdf, 'A',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    z_slice=0,
                                    rotation=quat,
                                    rot_origin=[0, 0, 0],
                                    normalize=False, hmin=False)
    img_xsecvec = interpolate_3d_cross_vec(sdf, 'A', 'B', 'C',
                                           x_pixels=50, y_pixels=50,
                                           xlim=(-1, 1), ylim=(-1, 1),
                                           z_slice=0,
                                           rotation=quat,
                                           rot_origin=[0, 0, 0],
                                           normalize=False, hmin=False)

    w_col = sdf['m'] / (sdf['rho'] * sdf['h'] ** 2)
    w_xsec = sdf['m'] / (sdf['rho'] * sdf['h'] ** 3)

    pos = quat.apply([1, 1, 1])
    val = quat.apply([sdf['A'][0], sdf['B'][0], sdf['C'][0]])
    real = -1 + (np.arange(0, 50) + 0.5) * (1 / 25)

    for y in range(50):
        for x in range(50):
            r = np.sqrt((pos[0] - real[x])**2 + (pos[1] - real[y])**2)
            w = column_kernel(r / sdf['h'][0], 3)

            assert img_col[y][x] == approx(w_col[0] * sdf['A'][0] * w)
            assert img_colvec[0][y][x] == approx(w_col[0] * val[0] * w)
            assert img_colvec[1][y][x] == approx(w_col[0] * val[1] * w)

            r = np.sqrt((pos[0] - real[x])**2
                        + (pos[1] - real[y])**2
                        + pos[2]**2)
            w = kernel.w(r / sdf['h'][0], 3)

            assert img_xsec[y][x] == approx(w_xsec[0] * sdf['A'][0] * w)

            assert img_xsecvec[0][y][x] == approx(w_xsec[0] * val[0] * w)
            assert img_xsecvec[1][y][x] == approx(w_xsec[0] * val[1] * w)

    img_grid = interpolate_3d_grid(sdf, 'A',
                                   x_pixels=50, y_pixels=50, z_pixels=50,
                                   xlim=(-1, 1), ylim=(-1, 1), zlim=(-1, 1),
                                   rotation=quat,
                                   rot_origin=[0, 0, 0],
                                   normalize=False, hmin=False)

    for z in range(50):
        for y in range(50):
            for x in range(50):
                r = np.sqrt((pos[0] - real[x])**2
                            + (pos[1] - real[y])**2
                            + (pos[2] - real[z])**2)
                w = kernel.w(r / sdf['h'][0], 3)
                grid_value = w_xsec[0] * sdf['A'][0] * w

                assert img_grid[z][y][x] == approx(grid_value)


@mark.parametrize("backend", backends)
def test_rotation_stability(backend: str) -> None:
    """
    Rotation should not change values at the rotation origin.
    """
    data = {'x': [1.0, 3.0], 'y': [1.0, -1.0], 'z': [1.0, -0.5],
            'A': [4.0, 3.0], 'B': [3.0, 2.0], 'C': [1.0, 1.5],
            'h': [0.9, 1.4], 'rho': [0.4, 0.6], 'm': [0.03, 0.06]}
    sdf = SarracenDataFrame(data, params=dict())
    kernel = CubicSplineKernel()
    sdf.kernel = kernel
    sdf.backend = backend

    real = -1 + (np.arange(0, 50) + 0.5) * (1 / 25)
    pixel_x, pixel_y = 12, 30

    functions_list: List[Callable] = [interpolate_3d_proj,
                                      interpolate_3d_cross]
    for func in functions_list:
        img = func(sdf, 'A',
                   x_pixels=50, y_pixels=50,
                   xlim=(-1, 1), ylim=(-1, 1),
                   normalize=False, hmin=False)
        img_rot = func(sdf, 'A',
                       x_pixels=50, y_pixels=50,
                       xlim=(-1, 1), ylim=(-1, 1),
                       rotation=[237, 0, 0],
                       rot_origin=[real[pixel_x], real[pixel_y], 0],
                       normalize=False, hmin=False)

        assert img[pixel_y][pixel_x] == approx(img_rot[pixel_y][pixel_x])


@mark.parametrize("backend", backends)
def test_axes_rotation_separation(backend: str) -> None:
    """
    Rotations should be independent of the defined x & y interpolation axes.
    Similar to test_image_transpose(), but a rotation is applied to all
    interpolations.
    """
    data = {'x': [-1, 1], 'y': [1, -1], 'z': [1, -1],
            'A': [2, 1.5], 'B': [2, 2], 'C': [4, 3],
            'h': [1.1, 1.3], 'rho': [0.55, 0.45], 'm': [0.04, 0.05]}
    sdf = SarracenDataFrame(data, params=dict())
    sdf.backend = backend

    rotation = [234, 90, 48]

    img1 = interpolate_3d_proj(sdf, 'A',
                               x_pixels=50,  y_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1),
                               rotation=rotation,
                               normalize=False, hmin=False)
    img2 = interpolate_3d_proj(sdf, 'A', x='y', y='x',
                               x_pixels=50,  y_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1),
                               rotation=rotation,
                               normalize=False, hmin=False)
    assert_allclose(img1, img2.T)

    img1_tuple = interpolate_3d_vec(sdf, 'A', 'B', 'C',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    rotation=rotation,
                                    normalize=False, hmin=False)
    img2_tuple = interpolate_3d_vec(sdf, 'A', 'B', 'C', x='y', y='x',
                                    x_pixels=50, y_pixels=50,
                                    xlim=(-1, 1), ylim=(-1, 1),
                                    rotation=rotation,
                                    normalize=False, hmin=False)
    assert_allclose(img1_tuple[0], img2_tuple[0].T)
    assert_allclose(img1_tuple[1], img2_tuple[1].T)

    img1 = interpolate_3d_cross(sdf, 'A',
                                x_pixels=50, y_pixels=50,
                                xlim=(-1, 1), ylim=(-1, 1),
                                z_slice=0,
                                rotation=rotation,
                                normalize=False, hmin=False)
    img2 = interpolate_3d_cross(sdf, 'A', x='y', y='x',
                                x_pixels=50, y_pixels=50,
                                xlim=(-1, 1), ylim=(-1, 1),
                                z_slice=0,
                                rotation=rotation,
                                normalize=False, hmin=False)
    assert_allclose(img1, img2.T)

    img1_tuple = interpolate_3d_cross_vec(sdf, 'A', 'B', 'C',
                                          x_pixels=50, y_pixels=50,
                                          xlim=(-1, 1), ylim=(-1, 1),
                                          z_slice=0,
                                          rotation=rotation,
                                          normalize=False, hmin=False)
    img2_tuple = interpolate_3d_cross_vec(sdf, 'A', 'B', 'C',
                                          x='y', y='x',
                                          x_pixels=50, y_pixels=50,
                                          xlim=(-1, 1), ylim=(-1, 1),
                                          z_slice=0,
                                          rotation=rotation,
                                          normalize=False, hmin=False)
    assert_allclose(img1_tuple[0], img2_tuple[0].T)
    assert_allclose(img1_tuple[1], img2_tuple[1].T)

    img1 = interpolate_3d_grid(sdf, 'A',
                               x_pixels=50, y_pixels=50, z_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1), zlim=(-1, 1),
                               rotation=rotation,
                               normalize=False, hmin=False)
    img2 = interpolate_3d_grid(sdf, 'A', x='y', y='x',
                               x_pixels=50, y_pixels=50, z_pixels=50,
                               xlim=(-1, 1), ylim=(-1, 1), zlim=(-1, 1),
                               rotation=rotation,
                               normalize=False, hmin=False)
    assert_allclose(img1, img2.transpose(0, 2, 1))


@mark.parametrize("backend", backends)
def test_axes_rotation_equivalency(backend: str) -> None:
    """
    A rotated interpolation (at multiples of 90 degrees) should be equivalent
    to a transformed interpolation with different x & y axes. For example,
    an interpolation rotated by 180 degrees around the z axis should be
    equivalent to the transpose of an unaltered interpolation.
    """
    data = {'x': [-1, 1], 'y': [1, -1], 'z': [1, -1],
            'A': [2, 1.5], 'B': [2, 2], 'C': [4, 3],
            'h': [1.1, 1.3], 'rho': [0.55, 0.45], 'm': [0.04, 0.05]}
    sdf = SarracenDataFrame(data, params=dict())
    sdf.backend = backend

    x, y, z = 'x', 'y', 'z'
    flip_x, flip_y, flip_z = False, False, False
    for i_z in range(4):
        for i_y in range(4):
            for i_x in range(4):
                rot_x, rot_y, rot_z = i_x * 90, i_y * 90, i_z * 90

                functions_list: List[Callable] = [interpolate_3d_proj,
                                                  interpolate_3d_cross]
                for func in functions_list:
                    img1 = func(sdf, 'A',
                                x_pixels=50, y_pixels=50,
                                xlim=(-1, 1), ylim=(-1, 1),
                                rotation=[rot_z, rot_y, rot_x],
                                rot_origin=[0, 0, 0],
                                normalize=False, hmin=False)
                    img2 = func(sdf, 'A',
                                x=x, y=y,
                                x_pixels=50, y_pixels=50,
                                xlim=(-1, 1), ylim=(-1, 1),
                                normalize=False, hmin=False)

                    img2 = img2 if not flip_x else np.flip(img2, 1)
                    img2 = img2 if not flip_y else np.flip(img2, 0)
                    assert_allclose(img1, img2)

                y, z = z, y
                flip_y, flip_z = not flip_z, flip_y
            x, z = z, x
            flip_x, flip_z = flip_z, not flip_x
        x, y = y, x
        flip_x, flip_y = not flip_y, flip_x


def test_com_rotation() -> None:
    """ Rotation around centre of mass should equal rotation around 0
    when positions have been reset around com. """

    x = np.array([0.80500292, 0.80794079, 0.51532556, 0.28580138, 0.0539307,
                  0.38336888, 0.40847321, 0.04527519, 0.04875771, 0.99917612])
    y = np.array([0.55559612, 0.2714516, 0.87965117, 0.06421444, 0.67918153,
                  0.8700885, 0.22731853, 0.89544824, 0.87219547, 0.01851722])
    z = np.array([0.32494264, 0.80621533, 0.31645209, 0.14903858, 0.69851199,
                  0.4485441, 0.79893949, 0.23551646, 0.31978465, 0.79987953])
    h = np.array([0.08582579, 0.08449268, 0.03678807, 0.09510229, 0.03994252,
                  0.09364420, 0.05561597, 0.02401353, 0.07414216, 0.06743897])
    val = np.array([3.9045891, 7.9793389, 3.8047537, 7.1325786, 6.125178,
                    9.4100098, 9.9167672, 7.2367625, 8.0884381, 1.5286502])
    mass = 3.2e-4

    sdf_com = SarracenDataFrame({'x': x, 'y': y, 'z': z, 'h': h, 'val': val},
                                params={'mass': mass, 'hfact': 1.2})

    com = [(x * mass).sum() / (10 * mass),
           (y * mass).sum() / (10 * mass),
           (z * mass).sum() / (10 * mass)]
    x = x - com[0]
    y = y - com[1]
    z = z - com[2]
    sdf_zero = SarracenDataFrame({'x': x, 'y': y, 'z': z, 'h': h, 'val': val},
                                 params={'mass': mass, 'hfact': 1.2})

    functions_list: List[Callable] = [interpolate_3d_proj,
                                      interpolate_3d_cross]
    for func in functions_list:
        img1 = func(sdf_com, 'val',
                    x_pixels=50, y_pixels=50,
                    rotation=[35, 60, 75],
                    rot_origin='com')
        img2 = func(sdf_zero, 'val',
                    x_pixels=50, y_pixels=50,
                    rotation=[35, 60, 75],
                    rot_origin=[0, 0, 0])

        assert_allclose(img1, img2)



"""pytest unit tests for kernel functionality."""
from typing import Callable
from pytest import approx, mark
from scipy.integrate import quad, dblquad, tplquad
import numpy as np

from sarracen.kernels import BaseKernel
from sarracen.kernels import CubicSplineKernel
from sarracen.kernels import QuarticSplineKernel
from sarracen.kernels import QuinticSplineKernel


def single_kernel(x: float, kernel: BaseKernel) -> float:
    return kernel.w(np.abs(x), 1)


def double_kernel(y: float, x: float, kernel: BaseKernel) -> float:
    # Utility function for double integrals in test_normalization
    return kernel.w(np.sqrt(x ** 2 + y ** 2), 2)


def triple_kernel(z: float, y: float, x: float, kernel: BaseKernel) -> float:
    # Utility function for triple integrals in test_normalization
    return kernel.w(np.sqrt(x ** 2 + y ** 2 + z ** 2), 3)


def double_column(y: float, x: float, column_func: Callable) -> float:
    return column_func(np.sqrt(x ** 2 + y ** 2), 0)


def test_cubicspline() -> None:
    kernel = CubicSplineKernel()

    # testing kernel values at q = 0
    # which should be equal to the normalization constants
    assert kernel.w(0, 1) == 2 / 3
    assert kernel.w(0, 2) == 10 / (7 * np.pi)
    assert kernel.w(0, 3) == 1 / np.pi

    # testing kernel values at q = 1
    assert kernel.w(1, 1) == 1 / 6
    assert kernel.w(1, 2) == 5 / (14 * np.pi)
    assert kernel.w(1, 3) == 1 / (4 * np.pi)

    # testing kernel values at q = 2
    assert kernel.w(2, 2) == 0
    assert kernel.w(10, 3) == 0


def test_quarticspline() -> None:
    kernel = QuarticSplineKernel()

    # unlike the cubic spline, these will NOT
    # be equal to the normalization constants.
    # kernel.w(0, d) = norm_d * 230/16
    assert kernel.w(0, 1) == approx(115 / 192)
    assert kernel.w(0, 2) == approx(1380 / (1199 * np.pi))
    assert kernel.w(0, 3) == approx(23 / (32 * np.pi))

    assert kernel.w(1, 1) == approx(19 / 96)
    assert kernel.w(1, 2) == approx(456 / (1199 * np.pi))
    assert kernel.w(1, 3) == approx(19 / (80 * np.pi))

    # these are equivalent to the normalization constants
    assert kernel.w(1.5, 1) == approx(1 / 24)
    assert kernel.w(1.5, 2) == approx(96 / (1199 * np.pi))
    assert kernel.w(1.5, 3) == approx(1 / (20 * np.pi))

    assert kernel.w(2.5, 2) == 0
    assert kernel.w(10, 3) == 0


def test_quinticspline() -> None:
    kernel = QuinticSplineKernel()

    # again, unlike the cubic spline, these will NOT
    # be equal to the normalization constants.
    # kernel.w(0, d) = norm_d * 66
    assert kernel.w(0, 1) == approx(11 / 20)
    assert kernel.w(0, 2) == approx(231 / (239 * np.pi))
    assert kernel.w(0, 3) == approx(11 / (20 * np.pi))

    assert kernel.w(1, 1) == approx(13 / 60)
    assert kernel.w(1, 2) == approx(91 / (239 * np.pi))
    assert kernel.w(1, 3) == approx(13 / (60 * np.pi))

    # these are equivalent to the normalization constants
    assert kernel.w(2, 1) == approx(1 / 120)
    assert kernel.w(2, 2) == approx(7 / (478 * np.pi))
    assert kernel.w(2, 3) == approx(1 / (120 * np.pi))

    assert kernel.w(3, 2) == 0
    assert kernel.w(10, 3) == 0


@mark.parametrize("kernel",
                  [CubicSplineKernel(),
                   QuarticSplineKernel(),
                   QuinticSplineKernel()])
def test_normalization(kernel: BaseKernel) -> None:

    # Since the three integrals below are only performed in positive space,
    # the resulting normalized values will not be equal to 1, rather 1/(2^dim).
    # This value represents the proportion of space in 1,2, and 3 dimensions
    # that has all positive coordinates.

    norm = quad(single_kernel, -kernel.get_radius(),
                kernel.get_radius(), tuple([kernel]))[0]
    assert approx(norm) == 1

    norm = dblquad(double_kernel, -kernel.get_radius(),
                   kernel.get_radius(), -kernel.get_radius(),
                   kernel.get_radius(), tuple([kernel]))[0]
    assert approx(norm) == 1

    norm = tplquad(triple_kernel, -kernel.get_radius(),
                   kernel.get_radius(), -kernel.get_radius(),
                   kernel.get_radius(), -kernel.get_radius(),
                   kernel.get_radius(), tuple([kernel]))[0]
    assert approx(norm) == 1


def test_cubic_column() -> None:
    kernel = CubicSplineKernel()
    column_kernel = kernel.get_column_kernel(10000)
    pts = np.linspace(0, kernel.get_radius(), 10000)
    # at q = 0, this integral is solvable analytically
    assert np.interp(0, pts, column_kernel) == approx(3 / (2 * np.pi))
    # numerically calculated values
    assert np.interp(0.5, pts, column_kernel) == approx(0.33875339978)
    assert np.interp(1, pts, column_kernel) == approx(0.111036060968)
    assert np.interp(1.5, pts, column_kernel) == approx(0.0114423169642)
    assert np.interp(2, pts, column_kernel) == 0
    assert np.interp(5, pts, column_kernel) == 0


def test_quartic_column() -> None:
    kernel = QuarticSplineKernel()
    column_kernel = kernel.get_column_kernel(10000)
    pts = np.linspace(0, kernel.get_radius(), 10000)

    # at q = 0, this integral is solvable analytically
    assert np.interp(0, pts, column_kernel) == approx(6 / (5 * np.pi))
    # numerically calculated values
    assert np.interp(0.5, pts, column_kernel) == approx(0.288815941868)
    assert np.interp(1, pts, column_kernel) == approx(0.120120735858)
    assert np.interp(1.5, pts, column_kernel) == approx(0.0233911861393)
    assert np.interp(2, pts, column_kernel) == approx(0.00116251851966)
    assert np.interp(2.5, pts, column_kernel) == 0
    assert np.interp(5, pts, column_kernel) == 0


def test_quintic_column() -> None:
    kernel = QuinticSplineKernel()
    column_kernel = kernel.get_column_kernel(10000)
    pts = np.linspace(0, kernel.get_radius(), 10000)

    # at q = 0, this integral is solvable analytically
    assert np.interp(0, pts, column_kernel) == approx(1 / np.pi)
    # numerically calculated values
    assert np.interp(0.5, pts, column_kernel) == approx(0.251567608959)
    assert np.interp(1, pts, column_kernel) == approx(0.121333261458)
    assert np.interp(1.5, pts, column_kernel) == approx(0.0328632154395)
    assert np.interp(2, pts, column_kernel) == approx(0.00403036583315)
    assert np.interp(2.5, pts, column_kernel) == approx(0.0000979416858548,
                                                        rel=1e-4)
    assert np.interp(3, pts, column_kernel) == 0
    assert np.interp(5, pts, column_kernel) == 0


@mark.parametrize("kernel",
                  [CubicSplineKernel(),
                   QuarticSplineKernel(),
                   QuinticSplineKernel()])
def test_normalized_column(kernel: BaseKernel) -> None:
    norm = dblquad(double_column, -kernel.get_radius(), kernel.get_radius(),
                   -kernel.get_radius(), kernel.get_radius(),
                   tuple([kernel.get_column_kernel_func(10000)]))[0]
    assert approx(norm) == 1


@mark.parametrize("kernel",
                  [CubicSplineKernel(),
                   QuarticSplineKernel(),
                   QuinticSplineKernel()])
def test_oob(kernel: BaseKernel) -> None:
    for dimensions in range(1, 3):
        assert kernel.w(-1, dimensions) == 0
        assert kernel.w(kernel.get_radius() + 1, dimensions) == 0

    column_func = kernel.get_column_kernel_func(1000)
    assert column_func(-1, 0) == column_func(0, 0)
    assert approx(column_func(kernel.get_radius() + 1, 0)) == 0



import numpy as np
from numba import njit

from ..kernels import BaseKernel


class CubicSplineKernel(BaseKernel):
    """An implementation of the Cubic Spline kernel"""

    @staticmethod
    def get_radius() -> float:
        return 2

    @staticmethod
    @njit(fastmath=True)
    def w(q: float, ndim: int) -> float:
        norm = 2 / 3 if (ndim == 1) \
            else 10 / (7 * np.pi) if (ndim == 2) \
            else 1 / np.pi

        return norm * ((1 - 1.5 * q**2 + 0.75 * q**3) * (0 <= q) * (q < 1)
                       + 0.25 * (2 - q)**3 * (1 <= q) * (q < 2))



from typing import Dict, Union, Any

import pandas as pd
import numpy as np
import sarracen
import tempfile
from pytest import fixture, mark, raises
from pandas import testing as tm

from sarracen import SarracenDataFrame

from sarracen.writers.write_phantom import (_check_for_essential_data,
                                            _standardize_dtypes,
                                            _validate_ntypes,
                                            _reorder_params,
                                            _validate_particle_counts,
                                            _validate_particle_masses)


@fixture
def particles_df() -> pd.DataFrame:
    x = [0, 0, 0, 0, 1, 1, 1, 1]
    y = [0, 0, 1, 1, 0, 0, 1, 1]
    z = [0, 1, 0, 1, 0, 1, 0, 1]
    h = [0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8]
    vx = [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    vy = [-0.2, -0.3, -0.4, -0.5, -0.6, -0.7, -0.8, -0.9]
    vz = [0.01, -0.01, 0.01, -0.01, 0.01, -0.01, 0.01, -0.01]

    return pd.DataFrame({'x': x, 'y': y, 'z': z, 'h': h,
                         'vx': vx, 'vy': vy, 'vz': vz})


def test_check_for_essential_data(particles_df: pd.DataFrame) -> None:
    """ Test that essential data is present in the dataframe."""

    sdf = SarracenDataFrame(particles_df)
    sdf = sdf.drop(columns=['h'])
    with raises(ValueError, match='columns'):
        _check_for_essential_data(sdf)

    sdf = SarracenDataFrame(particles_df)
    sdf = sdf.drop(columns=['z'])
    with raises(ValueError, match='columns'):
        _check_for_essential_data(sdf)


def test_params_dtype_standardization(particles_df: pd.DataFrame) -> None:
    """ Test that params dtype is correctly standardized."""

    params: Dict[str, Any] = {'massoftype': 1e-4,
                              'ntypes': 8,
                              'nparttot': 8,
                              'massoftype_7': 1e-6,
                              'file_identifier': 'test Phantom write'}

    params = _standardize_dtypes(params)

    assert isinstance(params['massoftype'], np.float64)
    assert isinstance(params['massoftype_7'], np.float64)
    assert isinstance(params['ntypes'], np.int32)
    assert isinstance(params['nparttot'], np.int32)
    assert isinstance(params['file_identifier'], str)


@mark.parametrize("id_method", ('ntypes', 'massoftype',
                                'npartoftype', 'itype'))
def test_validate_ntypes_12(particles_df: pd.DataFrame,
                            id_method: str) -> None:
    """ Test that number of particle types is correctly validated."""

    params: Dict[str, Union[np.generic, str]] = {}

    if id_method == 'ntypes':
        params['ntypes'] = np.int32(12)
    elif id_method == 'massoftype':
        params['massoftype'] = np.float64(1e-4)
        params['massoftype_7'] = np.float64(1e-6)
        params['massoftype_12'] = np.float64(1e-7)
    elif id_method == 'npartoftype':
        params['npartoftype'] = np.int32(10)
        params['npartoftype_7'] = np.int32(2)
        params['npartoftype_12'] = np.int32(6)
        params['npartoftype_13'] = np.int64(10)
        params['npartoftype_20'] = np.int64(2)
        params['npartoftype_24'] = np.int64(6)
    elif id_method == 'itype':
        particles_df['itype'] = [1, 1, 1, 1, 1, 7, 7, 12]

    sdf = sarracen.SarracenDataFrame(particles_df, params)

    params = _validate_ntypes(sdf, params)

    assert params['ntypes'] == 12


@mark.parametrize("id_method", ('base', 'ntypes', 'massoftype',
                                'npartoftype', 'itype'))
def test_validate_ntypes_8(particles_df: pd.DataFrame,
                           id_method: str) -> None:
    """ Test that number of particle types is correctly validated."""

    params: Dict[str, Union[np.generic, str]] = {}

    if id_method == 'ntypes':
        params['ntypes'] = np.int32(8)
    elif id_method == 'massoftype':
        params['massoftype'] = np.float64(1e-4)
        params['massoftype_7'] = np.float64(1e-6)
    elif id_method == 'npartoftype':
        params['npartoftype'] = np.int32(10)
        params['npartoftype_7'] = np.int32(2)
        params['npartoftype_12'] = np.int64(6)
        params['npartoftype_13'] = np.int64(10)
    elif id_method == 'itype':
        particles_df['itype'] = [1, 1, 1, 1, 1, 7, 7, 7]

    sdf = sarracen.SarracenDataFrame(particles_df, params)

    params = _validate_ntypes(sdf, params)

    assert params['ntypes'] == 8


@mark.parametrize("dust, id_method",
                  [(False, 'itype'), (False, 'npartoftype'),
                   (True, 'itype'), (True, 'npartoftype')])
def test_validate_particle_counts(particles_df: pd.DataFrame,
                                  dust: bool,
                                  id_method: str) -> None:
    """ Test that number of each particle type is correctly autofilled."""

    params: Dict[str,
                 Union[np.generic,
                       str]] = {'massoftype': np.float64(1e-4),
                                'ntypes': np.int32(8),
                                'file_identifier': 'test Phantom write'}

    # if npartoftype not present, then it determines based on itype alone
    if id_method == 'npartoftype':
        params['npartoftype'] = np.int32(5) if dust else np.int32(8)
        params['npartoftype_7'] = np.int32(3) if dust else np.int32(0)

    if dust:
        particles_df['itype'] = [1, 1, 1, 1, 1, 7, 7, 7]

    sdf = sarracen.SarracenDataFrame(particles_df, params)

    params = _validate_particle_counts(sdf, params)

    assert params['nparttot'] == 8
    assert params['npartoftype'] == 5 if dust else 8
    assert params['npartoftype_2'] == 0
    assert params['npartoftype_3'] == 0
    assert params['npartoftype_4'] == 0
    assert params['npartoftype_5'] == 0
    assert params['npartoftype_6'] == 0
    assert params['npartoftype_7'] == 3 if dust else 8
    assert params['npartoftype_8'] == 0
    assert params['nparttot_2'] == 8
    assert params['npartoftype_9'] == 5 if dust else 8
    assert params['npartoftype_10'] == 0
    assert params['npartoftype_11'] == 0
    assert params['npartoftype_12'] == 0
    assert params['npartoftype_13'] == 0
    assert params['npartoftype_14'] == 0
    assert params['npartoftype_15'] == 3 if dust else 8
    assert params['npartoftype_16'] == 0


def test_params_reordering(particles_df: pd.DataFrame) -> None:
    """ Test that disordered params keyes are written correctly."""

    params: Dict[str, Union[np.generic,
                            str]] = {'massoftype': np.float64(1e-4),
                                     'massoftype_7': np.float64(1e-6),
                                     'massoftype_3': np.float64(1e-3),
                                     'nparttot': np.int64(8),
                                     'nparttot_2': np.int32(8),
                                     'npartoftype_15': np.int64(3),
                                     'npartoftype_3': np.int32(4),
                                     'npartoftype': np.int32(1),
                                     'npartoftype_7': np.int32(3),
                                     'npartoftype_11': np.int64(4),
                                     'npartoftype_9': np.int64(1),
                                     'file_identifier': 'test Phantom write'}

    params = _reorder_params(params)

    keys = list(params.keys())

    assert keys.index('massoftype') < keys.index('massoftype_3')
    assert keys.index('massoftype_3') < keys.index('massoftype_7')
    assert keys.index('nparttot') < keys.index('nparttot_2')
    assert keys.index('npartoftype') < keys.index('npartoftype_3')
    assert keys.index('npartoftype_3') < keys.index('npartoftype_7')
    assert keys.index('npartoftype_7') < keys.index('npartoftype_9')
    assert keys.index('npartoftype_9') < keys.index('npartoftype_11')
    assert keys.index('npartoftype_11') < keys.index('npartoftype_15')

    assert params['massoftype'] == 1e-4
    assert params['massoftype_3'] == 1e-3
    assert params['massoftype_7'] == 1e-6
    assert params['nparttot'] == 8
    assert params['nparttot_2'] == 8
    assert params['npartoftype'] == 1
    assert params['npartoftype_3'] == 4
    assert params['npartoftype_7'] == 3
    assert params['npartoftype_9'] == 1
    assert params['npartoftype_11'] == 4
    assert params['npartoftype_15'] == 3


@mark.parametrize("id_method", ('params_mass', 'massoftype', 'mcol'))
def test_validate_particle_mass_gas(particles_df: pd.DataFrame,
                                    id_method: bool) -> None:
    """ Test that mass of each particle type is correctly validated."""

    params: Dict[str, Union[np.generic, str]] = {'ntypes': np.int32(8)}

    if id_method == 'params_mass':
        params['mass'] = np.float64(1e-4)
    elif id_method == 'massoftype':
        params['massoftype'] = np.float64(1e-4)
    elif id_method == 'mcol':
        particles_df['m'] = np.float64(1e-4)

    sdf = SarracenDataFrame(particles_df, params)

    params = _validate_particle_masses(sdf, params)

    assert params['massoftype'] == 1e-4
    assert params['massoftype_2'] == 0
    assert params['massoftype_3'] == 0
    assert params['massoftype_4'] == 0
    assert params['massoftype_5'] == 0
    assert params['massoftype_6'] == 0
    assert params['massoftype_7'] == 0
    assert params['massoftype_8'] == 0

    # with tempfile.NamedTemporaryFile() as fp:
    #     sarracen.write_phantom(fp.name, write_sdf)
    #     sdf = sarracen.read_phantom(fp.name)
    #
    #     assert isinstance(sdf, SarracenDataFrame)
    #     assert sdf.params is not None
    #     assert sdf.params['mass'] == 1e-4
    #     assert sdf.params['massoftype'] == 1e-4


@mark.parametrize("id_method", ('massoftype', 'mcol'))
def test_validate_particle_mass_dust(particles_df: pd.DataFrame,
                                     id_method: bool) -> None:
    """ Test that mass of each particle type is correctly validated."""

    params: Dict[str, Union[np.generic, str]] = {'ntypes': np.int32(8)}

    if id_method == 'massoftype':
        params['massoftype'] = np.float64(1e-4)
        params['massoftype_7'] = np.float64(1e-6)
    elif id_method == 'mcol':
        particles_df['m'] = [1e-4] * 5 + [1e-6] * 3

    sdf = SarracenDataFrame(particles_df, params)
    sdf['itype'] = [1, 1, 1, 1, 1, 7, 7, 7]

    params = _validate_particle_masses(sdf, params)

    assert params['massoftype'] == 1e-4
    assert params['massoftype_2'] == 0
    assert params['massoftype_3'] == 0
    assert params['massoftype_4'] == 0
    assert params['massoftype_5'] == 0
    assert params['massoftype_6'] == 0
    assert params['massoftype_7'] == 1e-6
    assert params['massoftype_8'] == 0


def test_write_gas(particles_df: pd.DataFrame) -> None:
    """ Test writing of simple gas-only particle dumpfile."""

    params = {'massoftype': np.float64(1e-4),
              'iexternalforce': np.int32(0),
              'udist': np.float64(2e-3),
              'utime': np.float64(2e-5),
              'umass': np.float64(2e-6),
              'umagfd': np.float64(2e-2),
              'file_identifier': 'test of Phantom writing'}

    write_sdf = SarracenDataFrame(particles_df, params)

    with tempfile.NamedTemporaryFile() as fp:
        sarracen.write_phantom(fp.name, write_sdf)
        sdf = sarracen.read_phantom(fp.name)

        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-4
        assert sdf.params['mass'] == 1e-4
        assert sdf.params['udist'] == 2e-3
        assert 'mass' in sdf.params
        assert 'mass' not in sdf.columns
        tm.assert_series_equal(sdf['x'], write_sdf['x'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['y'], write_sdf['y'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['z'], write_sdf['z'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['h'], write_sdf['h'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vx'], write_sdf['vx'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vy'], write_sdf['vy'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vz'], write_sdf['vz'],
                               check_index=False, check_names=False,
                               check_dtype=False)


def test_write_gas_and_dust(particles_df: pd.DataFrame) -> None:
    """ Test writing of simple gas and dust-only dumpfile."""

    params = {'massoftype': np.float64(1e-4),
              'massoftype_2': np.float64(0),
              'massoftype_3': np.float64(0),
              'massoftype_4': np.float64(0),
              'massoftype_5': np.float64(0),
              'massoftype_7': np.float64(1e-6),
              'iexternalforce': np.int32(0),
              'udist': np.float64(2e-3),
              'utime': np.float64(2e-5),
              'umass': np.float64(2e-6),
              'umagfd': np.float64(2e-2),
              'file_identifier': 'test of Phantom writing'}

    write_sdf = SarracenDataFrame(particles_df, params)

    write_sdf['itype'] = [1, 1, 1, 1, 1, 7, 7, 7]

    with tempfile.NamedTemporaryFile() as fp:
        sarracen.write_phantom(fp.name, write_sdf)

        sdf = sarracen.read_phantom(fp.name, separate_types=None)
        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-4
        assert sdf.params['massoftype_7'] == 1e-6
        assert sdf.params['udist'] == 2e-3
        assert 'mass' not in sdf.params
        assert 'mass' in sdf.columns

        mass = [1e-4] * 5 + [1e-6] * 3
        tm.assert_series_equal(sdf['mass'], pd.Series(mass),
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['itype'], write_sdf['itype'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['x'], write_sdf['x'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['y'], write_sdf['y'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['z'], write_sdf['z'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['h'], write_sdf['h'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vx'], write_sdf['vx'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vy'], write_sdf['vy'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vz'], write_sdf['vz'],
                               check_index=False, check_names=False,
                               check_dtype=False)


def test_write_gas_and_sink(particles_df: pd.DataFrame) -> None:
    """ Test writing of simple gas-only particle dumpfile."""

    params = {'massoftype': np.float64(1e-4),
              'iexternalforce': np.int32(0),
              'udist': np.float64(2e-3),
              'utime': np.float64(2e-5),
              'umass': np.float64(2e-6),
              'umagfd': np.float64(2e-2),
              'file_identifier': 'test of Phantom writing'}

    write_sdf = SarracenDataFrame(particles_df, params)

    sink_x = [0.5, 0.5]
    sink_y = [0.3, 0.3]
    sink_z = [0.1, 0.1]
    sink_m = [0.05, 0.001]
    sink_spinx = [-0.002, -0.002]
    sink_spiny = [1e-7, 1e-6]
    sink_spinz = [0.002, 0.002]

    sink_df = pd.DataFrame({'x': sink_x, 'y': sink_y, 'z': sink_z,
                            'm': sink_m, 'spinx': sink_spinx,
                            'spiny': sink_spiny, 'spinz': sink_spinz})

    write_sdf_sinks = SarracenDataFrame(sink_df, params)

    with tempfile.NamedTemporaryFile() as fp:
        sarracen.write_phantom(fp.name, write_sdf, write_sdf_sinks)
        sdf, sdf_sinks = sarracen.read_phantom(fp.name)

        assert isinstance(sdf, SarracenDataFrame)
        assert sdf.params is not None
        assert sdf.params['massoftype'] == 1e-4
        assert sdf.params['mass'] == 1e-4
        assert sdf.params['udist'] == 2e-3
        assert 'mass' in sdf.params
        assert 'mass' not in sdf.columns
        tm.assert_series_equal(sdf['x'], write_sdf['x'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['y'], write_sdf['y'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['z'], write_sdf['z'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['h'], write_sdf['h'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vx'], write_sdf['vx'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vy'], write_sdf['vy'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf['vz'], write_sdf['vz'],
                               check_index=False, check_names=False,
                               check_dtype=False)

        assert isinstance(sdf_sinks, SarracenDataFrame)
        tm.assert_series_equal(sdf_sinks['x'], write_sdf_sinks['x'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['y'], write_sdf_sinks['y'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['z'], write_sdf_sinks['z'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['m'], write_sdf_sinks['m'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spinx'], write_sdf_sinks['spinx'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spiny'], write_sdf_sinks['spiny'],
                               check_index=False, check_names=False,
                               check_dtype=False)
        tm.assert_series_equal(sdf_sinks['spinz'], write_sdf_sinks['spinz'],
                               check_index=False, check_names=False,
                               check_dtype=False)



import math

from numba import njit
from typing import Tuple


@njit
def line_int(r0: float, d1: float, d2: float, h: float) -> float:
    """ Calculate an exact 2D line integral over the cubic spline kernel.

    Used in exact calculation of a pixel surface integral in 2D.

    Parameters
    ----------
    r0: float
        Distance between the contributing particle and the line.
    d1, d2: float
        Distance from the endpoint of `r0` to each endpoint of the line.
    h: float
        Smoothing length of the contributing particle.

    Returns
    -------
    float
        The exact value of this line integral.
    """
    if r0 == 0:
        return 0
    elif r0 > 0:
        result = 1
        ar0 = r0
    else:
        result = -1
        ar0 = -r0

    q0 = ar0 / h

    # Determine the angle between q0 and the endpoints of the line,
    # relative to the contributing particle.
    phi1 = math.atan(abs(d1) / ar0)
    phi2 = math.atan(abs(d2) / ar0)

    if d1 * d2 >= 0:
        # Both line endpoints are on opposite sides of r0.
        result = result * (_full_2d_mod(phi1, q0) + _full_2d_mod(phi2, q0))
    elif abs(d1) < abs(d2):
        # Both line endpoints are on the same side of r0,
        # with d2 having a larger magnitude.
        result = result * (_full_2d_mod(phi2, q0) - _full_2d_mod(phi1, q0))
    else:
        # Both line endpoints are on the same side of r0,
        # with d1 having a larger magnitude.
        result = result * (_full_2d_mod(phi1, q0) - _full_2d_mod(phi2, q0))

    return result


@njit
def _full_2d_mod(phi: float, q0: float) -> float:
    """ Calculate an exact 2D line integral over the cubic spline kernel.

    Assumes that one endpoint of the line is at the end of `q0`. Used in pint.

    Parameters
    ----------
    phi: float
        Angle between `q0` and the endpoint of the line, relative to the
        contributing particle.
    q0: float
        The distance between the contributing particle and the line, scaled by
        the smoothing length of the particle.

    Returns
    -------
    float
        The exact value of this line integral.
    """
    if q0 <= 1.0:
        # At least part of the line lies within 0 < q <= 1
        q = q0 / math.cos(phi)

        if q <= 1.0:
            # The line lies entirely within 0 < q <= 1.
            return _f1_2d(phi, q0)
        elif q <= 2.0:
            # The line lies partly in 0 < q <= 1 and partly in 1 < q <= 2.

            # Angle between q0 and the line region endpoint within 0 < q <= 1,
            # relative to the contributing particle.
            phi1 = math.acos(q0)
            return _f2_2d(phi, q0) - _f2_2d(phi1, q0) + _f1_2d(phi1, q0)
        else:
            # The line spans all three possible regions, 0 < q <= 1,
            # 1 < q <= 2, and q > 2.

            # Angle between q0 and the line region endpoint within 0 < q <= 1,
            # relative to the contributing particle.
            phi1 = math.acos(q0)
            # Angle between q0 and the line region endpoint within 1 < q <= 2,
            # relative to the contributing particle.
            phi2 = math.acos(0.5 * q0)
            return _f3_2d(phi) - _f3_2d(phi2) + _f2_2d(phi2, q0) \
                - _f2_2d(phi1, q0) + _f1_2d(phi1, q0)
    elif q0 <= 2.0:
        # No part of the line lies within 0 < q <= 1, but it does lie within
        # 1 < q <= 2.
        q = q0 / math.cos(phi)

        if q <= 2.0:
            # The line lies entirely within 1 < q <= 2.
            return _f2_2d(phi, q0)
        else:
            # The line lies partly in 1 < q <= 2 and q > 2.

            # Angle between q0 and the line region endpoint within 1 < q <= 2,
            # relative to the contributing particle.
            phi2 = math.acos(0.5 * q0)
            return _f3_2d(phi) - _f3_2d(phi2) + _f2_2d(phi2, q0)
    else:
        # The line lies entirely within q > 2.
        return _f3_2d(phi)


@njit
def _f1_2d(phi: float, q0: float) -> float:
    """ Calculate an exact 2D line integral over the cubic spline kernel.

    Assumes that one endpoint of the line is at the end of `q0`. Only valid
    for 0 < q <= 1.
    Used in _full_2d_mod.

    Parameters
    ----------
    phi: float
        Angle between `q0` and the endpoint of the line segment, relative to
        the contributing particle.
    q0: float
        The distance between the contributing particle and the line, scaled by
        the smoothing length of the particle.

    Returns
    -------
    float
        The exact value of this line integral segment.
    """
    cphi2 = math.cos(phi) ** 2
    logs = math.log(math.tan(phi / 2. + math.pi / 4.))

    i2 = math.tan(phi)
    i4 = 1. / 3. * math.tan(phi) * (2. + 1. / cphi2)
    i5 = 1. / 16. * (0.5 * (11. * math.sin(phi) + 3. * math.sin(3. * phi))
                     / cphi2 / cphi2 + 6. * logs)

    return 5. / 7. * q0**2 / math.pi \
        * (i2 - 0.75 * q0**2 * i4 + 0.3 * q0**3 * i5)


@njit
def _f2_2d(phi: float, q0: float) -> float:
    """ Calculate an exact 2D line integral over the cubic spline kernel.

    Assumes that one endpoint of the line is at the end of `q0`. Only valid
    for 1 < q <= 2. Used in _full_2d_mod.

    Parameters
    ----------
    phi: float
        Angle between `q0` and the endpoint of the line segment, relative to
        the contributing particle.
    q0: float
        The distance between the contributing particle and the line, scaled by
        the smoothing length of the particle.

    Returns
    -------
    float
        The exact value of this line integral segment.
    """
    cphi2 = math.cos(phi) ** 2

    q02 = q0 * q0
    q03 = q02 * q0

    logs = math.log(math.tan(phi / 2. + math.pi / 4.))

    i0 = phi
    i2 = math.tan(phi)
    i3 = 1. / 2. * (math.tan(phi) / math.cos(phi) + logs)
    i4 = 1. / 3. * math.tan(phi) * (2. + 1. / cphi2)
    i5 = 1. / 16. * (0.5 * (11. * math.sin(phi)
                            + 3. * math.sin(3. * phi)) / cphi2 / cphi2
                     + 6. * logs)

    return 5. / 7. * q02 / math.pi * (
            2. * i2 - 2. * q0 * i3 + 3. / 4. * q02 * i4 - 1. / 10. * q03 * i5
            - 1. / 10. / q02 * i0)


@njit
def _f3_2d(phi: float) -> float:
    """ Calculate an exact 2D line integral over the cubic spline kernel.

    Assumes that one endpoint of the line is at the end of `q0`. Only valid
    for q > 2. Used in _full_2d_mod.

    Parameters
    ----------
    phi: float
        Angle pointing towards the endpoint of the line segment, relative to
        the contributing particle.

    Returns
    -------
    float
        The exact value of this line integral segment.
    """
    return 0.5 / math.pi * phi


@njit
def surface_int(r0: float,
                x1: float,
                y1: float,
                x2: float,
                y2: float,
                wx: float,
                wy: float,
                h: float) -> float:
    """ Calculate an exact 3D surface integral over the cubic spline kernel.

    Used to exactly calculating the contribution of a particle to a pixel's
    volume in 3D space.

    Parameters
    ----------
    r0: float
        Distance between the contributing particle and the target surface.
    x1, y1, x2, y2: float
        Upper and lower bounds of the target surface, relative to r0.
    wx, wy: float
        The size of a single pixel.
    h: float
        The smoothing length of the contributing particle.

    Returns
    -------
    float
        The exact value of this surface integral.
    """
    result = 0.0
    dx = x2 - x1
    dy = y2 - y1

    # Calculate the exact value of this surface by summing the comprising line
    # integrals.

    # Bottom boundary
    r1 = 0.5 * wy + dy
    d1 = 0.5 * wx - dx
    d2 = 0.5 * wx + dx
    result = result + _line_int3d(r0, r1, d1, d2, h)

    # Top boundary
    r1 = 0.5 * wy - dy
    d1 = 0.5 * wx + dx
    d2 = 0.5 * wx - dx
    result = result + _line_int3d(r0, r1, d1, d2, h)

    # Right boundary
    r1 = 0.5 * wx + dx
    d1 = 0.5 * wy + dy
    d2 = 0.5 * wy - dy
    result = result + _line_int3d(r0, r1, d1, d2, h)

    # Left boundary
    r1 = 0.5 * wx - dx
    d1 = 0.5 * wy - dy
    d2 = 0.5 * wy + dy
    result = result + _line_int3d(r0, r1, d1, d2, h)

    return result


@njit
def _line_int3d(r0: float, r1: float, d1: float, d2: float, h: float) -> float:
    """ Calculate an exact 3D line integral over the cubic spline kernel.

    Used in wallint.

    Parameters
    ----------
    r0: float
        Distance between the contributing particle and the target surface.
    r1: float
        Distance between the endpoint of `r0` and the line.
    d1, d2: float
        Distance from the endpoint of `r1` to each endpoint of the line.
    h: float
        Smoothing length of the contributing particle.

    Returns
    -------
    float
        The exact value of this line integral.
    """
    if abs(r0) == 0:
        return 0
    if r0 > 0:
        result = 1
        ar0 = r0
    else:
        result = -1
        ar0 = -r0

    if r1 > 0.:
        ar1 = r1
    else:
        result = -result
        ar1 = -r1

    # Split this line integral into two separate line integrals,
    # where one end point is at the endpoint of r1,
    # and the other end point is d1 or d2 respectively.
    int1 = _full_integral_3d(d1, ar0, ar1, h)
    int2 = _full_integral_3d(d2, ar0, ar1, h)

    if int1 < 0:
        int1 = 0
    if int2 < 0:
        int2 = 0

    if d1 * d2 >= 0:
        # Both line endpoints are on opposite sides of r1.
        result = result * (int1 + int2)
        if int1 + int2 < 0:
            print('Error: int1 + int2 < 0')
    elif abs(d1) < abs(d2):
        # Both line endpoints are on the same side of r1,
        # with d2 having a larger magnitude.
        result = result * (int2 - int1)
        if int2 - int1 < 0:
            print('Error: int2 - int1 < 0: ', int1, int2, '(', d1, d2, ')')
    else:
        # Both line endpoints are on the same side of r1,
        # with d1 having a larger magnitude.
        result = result * (int1 - int2)
        if int1 - int2 < 0:
            print('Error: int1 - int2 < 0: ', int1, int2, '(', d1, d2, ')')

    return result


@njit
def _full_integral_3d(d: float, r0: float, r1: float, h: float) -> float:
    """ Calculate an exact 3D line integral over the cubic spline kernel.

    Assumes that one endpoint of the line is at the end of `r1`.
    Used in _pint3d.

    Parameters
    ----------
    d: float
        Distance of the line.
    r0: float
        Distance between the target surface and the contributing particle.
    r1: float
        Distance between the end of `r0` and the line.
    h: float
        The smoothing length of the contributing particle.

    Returns
    -------
    float
        The exact value of this line integral.
    """
    r0h = r0 / h
    # Angle between the end of the line and the end of r1,
    # relative to the start of r1.
    phi = math.atan(abs(d) / r1)

    if abs(r0h) == 0 or abs(r1 / h) == 0 or abs(phi) == 0:
        return 0

    h2 = h * h
    r03 = r0 * r0 * r0
    r0h2 = r0h * r0h
    r0h3 = r0h2 * r0h
    r0h_2 = 1. / r0h2
    r0h_3 = 1. / r0h3

    b1 = 0.0
    b2 = 0.0
    if r0 > 2.0 * h:
        # The entire surface lies outside r > 2h.
        b3 = 0.25 * h2 * h
    elif r0 > h:
        # A part of the surface lies in the region h < r <= 2h.
        b3 = 0.25 * r03 * (-4. / 3. + r0h - 0.3 * r0h2 + 1. / 30. * r0h3
                           - 1. / 15. * r0h_3 + 8. / 5. * r0h_2)
        b2 = 0.25 * r03 * (-4. / 3. + r0h - 0.3 * r0h2 + 1. / 30. * r0h3
                           - 1. / 15. * r0h_3)
    else:
        # A part of the surface lies in the region 0 < r <= h.
        b3 = 0.25 * r03 * (-2. / 3. + 0.3 * r0h2 - 0.1 * r0h3 + 1.4 * r0h_2)
        b2 = 0.25 * r03 * (-2. / 3. + 0.3 * r0h2 - 0.1 * r0h3 - 0.2 * r0h_2)
        b1 = 0.25 * r03 * (-2. / 3. + 0.3 * r0h2 - 0.1 * r0h3)

    a = r1 / r0
    a2 = a * a

    # Squared distance between the contributing particle and the end of r1.
    linedist2 = r0 * r0 + r1 * r1
    # Distance between the end of r1 and the end of the line.
    r_ = r1 / math.cos(phi)
    # Squared distance between the contributing particle and the end of the
    # line.
    r2 = (r0 * r0 + r_ * r_)

    d2 = 0.0
    d3 = 0.0

    if linedist2 < h2:
        # A portion of the line lies within 0 < r < h.
        i = get_I_terms(r1 / math.sqrt(h2 - r0 * r0), a2, a)

        d2 = -1. / 6. * i[2] + 0.25 * r0h * i[3] - 0.15 * r0h2 * i[4]
        d2 += 1. / 30. * r0h3 * i[5] - 1. / 60. * r0h_3 * i[1]
        d2 += (b1 - b2) / r03 * i[0]

    if linedist2 < 4. * h2:
        # A portion of the line lies within 0 < r < 2h.
        i = get_I_terms(r1 / math.sqrt(4.0 * h2 - r0 * r0), a2, a)

        d3 = 1. / 3. * i[2] - 0.25 * r0h * i[3] + 3. / 40. * r0h2 * i[4]
        d3 += -1. / 120. * r0h3 * i[5] + 4. / 15. * r0h_3 * i[1]
        d3 += (b2 - b3) / r03 * i[0] + d2

    i = get_I_terms(math.cos(phi), a2, a)

    if r2 <= h2:
        # The entire line lies within 0 < r <= h.
        return r0h3 / math.pi * (1. / 6. * i[2] - 3. / 40. * r0h2 * i[4]
                                 + 1. / 40. * r0h3 * i[5] + b1 / r03 * i[0])
    elif r2 <= 4. * h2:
        # The entire line lies within 0 < r <= 2h.
        return r0h3 / math.pi * (0.25 * (4. / 3. * i[2] - (r0 / h) * i[3]
                                         + 0.3 * r0h2 * i[4]
                                         - 1. / 30. * r0h3 * i[5] +
                                         1. / 15. * r0h_3 * i[1])
                                 + b2 / r03 * i[0] + d2)
    else:
        # The line lies in all possible regions, 0 < r <= h, 0 < r <= 2h,
        # and r > 2h.
        return r0h3 / math.pi * (-0.25 * r0h_3 * i[1] + b3 / r03 * i[0] + d3)


@njit
def get_I_terms(cosp: float,
                a2: float,
                a: float) -> Tuple[float, float, float, float, float, float]:
    """Calculate I constants for calculations in _full_integral_3d"""
    cosp2 = cosp * cosp
    p = math.acos(cosp)
    tamath = math.sqrt(1. - cosp2) / cosp

    mu2_1 = 1. / (1. + cosp2 / a2)
    I0 = p
    I_2 = p + a2 * tamath
    I_4 = p + 2. * a2 * tamath + 1. / 3. * a2 * a2 * tamath * (2. + 1. / cosp2)

    u2 = (1. - cosp2) * mu2_1
    u = math.sqrt(u2)
    logs = math.log((1. + u) / (1. - u))
    I1 = math.atan2(u, a)

    fac = 1. / (1. - u2)
    I_1 = 0.5 * a * logs + I1
    I_3 = I_1 + a * 0.25 * (1. + a2) * (2. * u * fac + logs)

    I_5 = I_3 + (a * (1. + a2)**2 / 16. *
                 ((10. * u - 6. * u * u2) * fac**2 + 3. * logs))

    return I0, I1, I_2, I_3, I_4, I_5



import numpy as np
from numpy.testing import assert_array_equal
from sarracen import SarracenDataFrame
from sarracen.disc import angular_momentum


def test_mass_equivalency() -> None:
    """ Column mass result should equal global params mass result. """

    # randomly place particles
    rng = np.random.default_rng(seed=5)
    x = rng.random(100)
    y = rng.random(100)
    z = rng.random(100)
    vx = rng.random(100)
    vy = rng.random(100)
    vz = rng.random(100)
    mass = [3.2e-4] * 100

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z,
                                  'vx': vx, 'vy': vy, 'vz': vz,
                                  'mass': mass})
    Lx1, Ly1, Lz1 = angular_momentum(sdf)

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z,
                                  'vx': vx, 'vy': vy, 'vz': vz},
                            params={'mass': 3.2e-4})
    Lx2, Ly2, Lz2 = angular_momentum(sdf)

    assert_array_equal(Lx1, Lx2)
    assert_array_equal(Ly1, Ly2)
    assert_array_equal(Lz1, Lz2)


def test_parts_vs_whole() -> None:
    """ Profiles should be the same for matching bins. """

    # randomly place particles
    rng = np.random.default_rng(seed=5)
    x = rng.random(100)
    y = rng.random(100)
    z = rng.random(100)
    vx = rng.random(100)
    vy = rng.random(100)
    vz = rng.random(100)
    mass = [3.2e-4] * 100

    sdf = SarracenDataFrame(data={'x': x, 'y': y, 'z': z,
                                  'vx': vx, 'vy': vy, 'vz': vz,
                                  'mass': mass})
    Lx_in, Ly_in, Lz_in = angular_momentum(sdf, r_in=0.0, r_out=0.5, bins=100)
    Lx_ex, Ly_ex, Lz_ex = angular_momentum(sdf, r_in=0.5, r_out=1.0, bins=100)
    Lx, Ly, Lz = angular_momentum(sdf, r_in=0.0, r_out=1.0, bins=200)

    assert_array_equal(Lx_in, Lx[:100])
    assert_array_equal(Ly_in, Ly[:100])
    assert_array_equal(Lz_in, Lz[:100])
    assert_array_equal(Lx_ex, Lx[100:])
    assert_array_equal(Ly_ex, Ly[100:])
    assert_array_equal(Lz_ex, Lz[100:])



"""
Contains several interpolation functions which produce interpolated 2D or 1D
arrays of SPH data.
"""
import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation

from ..interpolate import BaseBackend, CPUBackend, GPUBackend
from ..kernels import BaseKernel

from typing import Tuple, Union, Optional, Type, Literal


def _default_xy(data: 'SarracenDataFrame',  # noqa: F821
                x: Union[str, None],
                y: Union[str, None]) -> Tuple[str, str]:
    """
    Utility function to determine the x & y columns to use during 2D
    interpolation.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y: str
        The x and y directional column labels passed to the interpolation
        function.

    Returns
    -------
    x, y: str
        The directional column labels to use in interpolation.
    """
    if x is None:
        x = data.xcol if not y == data.xcol else data.ycol
    if y is None:
        y = data.ycol if not x == data.ycol else data.xcol

    return x, y


def _default_xyz(data: 'SarracenDataFrame',  # noqa: F821
                 x: Union[str, None],
                 y: Union[str, None],
                 z: Union[str, None]) -> Tuple[str, str, str]:
    """
    Utility function to determine the x, y and z columns to use during 3-D
    interpolation.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y, z: str
        The x, y and z directional column labels passed to the interpolation
        function.

    Returns
    -------
    x, y, z: str
        The directional column labels to use in interpolation.
    """
    xcol = data.xcol
    ycol = data.ycol
    zcol = data.zcol

    if x is None:
        x = xcol if not y == xcol and not z == xcol else \
            ycol if not y == ycol and not z == ycol else zcol
    if y is None:
        y = ycol if not x == ycol and not z == ycol else \
            xcol if not x == xcol and not z == xcol else zcol
    if z is None:
        z = zcol if not x == zcol and not y == zcol else \
            ycol if not x == ycol and not y == ycol else xcol

    return x, y, z


def _default_bounds(x_data: np.ndarray,
                    y_data: np.ndarray,
                    xlim: Union[Tuple[Union[float, None], Union[float, None]],
                                None],
                    ylim: Union[Tuple[Union[float, None], Union[float, None]],
                                None]) -> Tuple[Tuple[float, float],
                                                Tuple[float, float]]:
    """
    Utility function to determine the 2-dimensional boundaries to use in 2D
    interpolation.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y: str
        The directional column labels that will be used in interpolation.
    x_data, y_data, z_data: ndarray
        The particle coordinate data.
    xlim, ylim: tuple of float
        The minimum and maximum values passed to the interpolation function, in
        particle data space.

    Returns
    -------
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the maximum and minimum values of `x` and `y`,
        snapped to the nearest integer.
    """
    # boundaries of the plot default to the max & min values of the data.
    x_min = xlim[0] if xlim is not None and xlim[0] is not None else None
    y_min = ylim[0] if ylim is not None and ylim[0] is not None else None
    x_max = xlim[1] if xlim is not None and xlim[1] is not None else None
    y_max = ylim[1] if ylim is not None and ylim[1] is not None else None

    x_min = x_data.min() if x_min is None else x_min
    y_min = y_data.min() if y_min is None else y_min
    x_max = x_data.max() if x_max is None else x_max
    y_max = y_data.max() if y_max is None else y_max

    return (x_min, x_max), (y_min, y_max)


def _set_pixels(x_pixels: Union[int, None],
                y_pixels: Union[int, None],
                xlim: Tuple[float, float],
                ylim: Tuple[float, float]) -> Tuple[int, int]:
    """
    Utility function to determine the number of pixels to interpolate over in
    2D interpolation.

    Parameters
    ----------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions passed to the
        interpolation function.
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle
        data space.

    Returns
    -------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions to use in 2D
        interpolation.
    """
    # set # of pixels to maintain an aspect ratio that is the same as the
    # underlying bounds of the data.

    dx = xlim[1] - xlim[0]
    dy = ylim[1] - ylim[0]

    if y_pixels is None:
        if x_pixels is None:
            x_pixels = 512
        y_pixels = int(np.rint(x_pixels * (dy / dx)))
    elif x_pixels is None:
        x_pixels = int(np.rint(y_pixels * (dx / dy)))

    return x_pixels, y_pixels


def _verify_columns(data: 'SarracenDataFrame',  # noqa: F821
                    x: str,
                    y: str) -> None:
    """
    Verify that columns required for 2D interpolation exist in `data`.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y: str
        The directional column labels that will be used in interpolation.

    Raises
    ------
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do
        not exist in `data`.
    """
    if x not in data.columns:
        raise KeyError(f"x-directional column '{x}' does not exist in the "
                       f"provided dataset.")
    if y not in data.columns:
        raise KeyError(f"y-directional column '{y}' does not exist in the "
                       f"provided dataset.")
    if data.hcol is None:
        raise KeyError("Smoothing length column does not exist in the "
                       "provided dataset.")
    if data.mcol is None and 'mass' not in data.params:
        raise KeyError("Missing particle mass data in this "
                       "SarracenDataFrame.")
    if data.rhocol is None and 'hfact' not in data.params:
        raise KeyError("Density cannot be derived from the columns in "
                       "this SarracenDataFrame.")


def _check_boundaries(x_pixels: int,
                      y_pixels: int,
                      xlim: Tuple[float, float],
                      ylim: Tuple[float, float]) -> None:
    """
    Verify that the pixel count and boundaries of a 2D plot describe a valid
    region.

    Parameters
    ----------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions passed to the
        interpolation function.
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle
        data space.

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximum values result in an invalid
        region.
    """
    if xlim[1] - xlim[0] <= 0:
        raise ValueError("`xlim` max must be greater than min!")
    if ylim[1] - ylim[0] <= 0:
        raise ValueError("`ylim` max must be greater than min!")
    if x_pixels <= 0:
        raise ValueError("`x_pixels` must be greater than zero!")
    if y_pixels <= 0:
        raise ValueError("`y_pixels` must be greater than zero!")


def _check_dimension(data: 'SarracenDataFrame',  # noqa: F821
                     dim: Literal[2, 3]) -> None:
    """
    Verify that a given dataset describes data with a required number of
    dimensions.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    dim: [2, 3]
        The number of required dimensions.

    Raises
    ------
    ValueError
        If the dataset is not `dim`-dimensional or `dim` is not 2 or 3.
    """
    if dim not in [2, 3]:
        raise ValueError("`dim` must be 2 or 3.")
    if data.get_dim() != dim:
        raise ValueError(f"Dataset is not {dim}-dimensional.")


def _rotate_data(data: 'SarracenDataFrame',  # noqa: F821
                 x_data: np.ndarray,
                 y_data: np.ndarray,
                 z_data: np.ndarray,
                 rotation: Union[np.ndarray, list, Rotation, None],
                 rot_origin: Union[np.ndarray, list, pd.Series,
                                   str, None]) -> Tuple[np.ndarray,
                                                        np.ndarray,
                                                        np.ndarray]:
    """
    Rotate vector data in a particle dataset.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x_data, y_data, z_data: ndarray
        The directional vector data.
    rotation: array_like or SciPy Rotation
        The rotation to apply to the vector data. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint']
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.

    Returns
    -------
    x_data, y_data, z_data: ndarray
        The rotated x, y, and z directional data.
    """
    vectors = np.array([x_data, y_data, z_data]).transpose()
    if rotation is not None:
        if not isinstance(rotation, Rotation):
            rotation_obj = Rotation.from_euler('zyx',
                                               rotation,
                                               degrees=True)
        else:
            rotation_obj = rotation

        # default
        if rot_origin is None:
            rot_origin_arr = np.array([0, 0, 0], dtype=x_data.dtype)

        # built-in point
        elif isinstance(rot_origin, str):
            if rot_origin == 'com':
                rot_origin_arr = np.asarray(data.centre_of_mass(),
                                            dtype=x_data.dtype)
            elif rot_origin == 'midpoint':
                rot_origin_arr = (vectors.min(0) + vectors.max(0)) / 2
            else:
                raise ValueError("rot_origin should be an [x, y, z] point or "
                                 "'com' or 'midpoint'")

        # user specified point
        elif isinstance(rot_origin, (list, pd.Series, np.ndarray)):
            if len(rot_origin) != 3:
                raise ValueError("rot_origin should specify [x, y, z] point.")
            rot_origin_arr = np.asarray(rot_origin, dtype=x_data.dtype)

        else:
            raise ValueError("rot_origin should be an [x, y, z] point or "
                             "'com' or 'midpoint'")

        vectors = vectors - rot_origin_arr
        vectors = rotation_obj.apply(vectors)
        vectors = vectors + rot_origin_arr

    return vectors[:, 0], vectors[:, 1], vectors[:, 2]


def _rotate_xyz(data: 'SarracenDataFrame',  # noqa: F821
                x: str,
                y: str,
                z: str,
                rotation: Union[np.ndarray, list, Rotation, None],
                rot_origin: Union[np.ndarray, list,
                                  str, None]) -> Tuple[np.ndarray,
                                                       np.ndarray,
                                                       np.ndarray]:
    """
    Rotate positional data in a particle dataset.

    Differs from _rotate_data() in that the returned data values are shuffled
    to ensure that the rotation is always applied to the global x, y, and z
    columns of the dataset, no matter the order of x, y, and z provided to
    this function.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y, z: str
        Directional column labels containing the positional column labels.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.

    Returns
    -------
    x_data, y_data, z_data: ndarray
        The rotated x, y, and z directional data.
    """
    rotated_x, rotated_y, rotated_z = _rotate_data(data, data[data.xcol],
                                                   data[data.ycol],
                                                   data[data.zcol], rotation,
                                                   rot_origin)
    x_data = rotated_x if x == data.xcol else \
        rotated_y if x == data.ycol else \
        rotated_z if x == data.zcol else data[x]
    y_data = rotated_x if y == data.xcol else \
        rotated_y if y == data.ycol else \
        rotated_z if y == data.zcol else data[y]
    z_data = rotated_x if z == data.xcol else \
        rotated_y if z == data.ycol else \
        rotated_z if z == data.zcol else data[z]

    return x_data, y_data, z_data


def _corotate(corotation: Union[np.ndarray, list],
              rotation: Union[np.ndarray, list,
                              Rotation, None]) -> Tuple[Union[np.ndarray,
                                                              list,
                                                              Rotation],
                                                        Union[np.ndarray,
                                                              list]]:
    """
    Calculates the rotation matrix for a corotating frame.

    Parameters
    ----------
    corotation: array_like
        The x, y, z coordinates of two locations which determines the
        corotating frame. Each coordinate is also array_like.
    rotation: array_like or SciPy Rotation
        An additional rotation to apply to the corotating frame.

    Returns
    -------
    rotation: array_like or SciPy Rotation
        The rotation to apply to the data before interpolation.
    rot_origin: array_like
        Point of rotation of the data.
    """
    corotation[1][0] -= corotation[0][0]
    corotation[1][1] -= corotation[0][1]
    corotation[1][2] -= corotation[0][2]

    rot_origin = corotation[0]
    angle = -np.arctan2(corotation[1][1], corotation[1][0])
    if rotation is None:
        rotation = np.array([angle * 180 / np.pi, 0, 0])
    else:
        if isinstance(rotation, Rotation):
            rotation = rotation.as_rotvec(degrees=True)
        rotation = np.array([angle * 180/np.pi + rotation[0],
                             rotation[1],
                             rotation[2]])
        rotation = Rotation.from_euler('zyx', rotation, degrees=True)

    return rotation, rot_origin


def _get_mass(data: 'SarracenDataFrame') -> Union[np.ndarray,  # noqa: F821
                                                  float]:
    if data.mcol is None:
        return data.params['mass']

    return data[data.mcol].to_numpy()


def _get_density(data: 'SarracenDataFrame') -> np.ndarray:  # noqa: F821
    if data.rhocol is None:
        hfact = data.params['hfact']
        mass = _get_mass(data)
        return ((hfact / data[data.hcol])**(data.get_dim()) * mass).to_numpy()

    return data[data.rhocol].to_numpy()


def _get_weight(data: 'SarracenDataFrame',  # noqa: F821
                target: Union[str, np.ndarray],
                dens_weight: bool) -> np.ndarray:

    if type(target) is str:
        if target == 'rho':
            target_data = _get_density(data)
        else:
            if target not in data.columns:
                raise KeyError(f"Target column '{target}' does not exist in "
                               f"provided dataset.")

            target_data = data[target].to_numpy()
    elif type(target) is np.ndarray:
        target_data = target
    else:
        raise KeyError(f"Target must be of type str or ndarray. "
                       f"Found: '{type(target)}'")

    mass_data = _get_mass(data)
    if dens_weight:
        return target_data * mass_data
    else:
        rho_data = _get_density(data)
        return target_data * mass_data / rho_data


def _get_smoothing_lengths(data: 'SarracenDataFrame',  # noqa: F821
                           hmin: bool,
                           x_pixels: int,
                           y_pixels: int,
                           xlim: Tuple[float, float],
                           ylim: Tuple[float, float]) -> np.ndarray:
    """ Return smoothing lengths, imposing a min length if hmin is True. """

    if hmin:
        pix_size = (xlim[1] - xlim[0]) / x_pixels
        pix_size = np.maximum(pix_size, (ylim[1] - ylim[0]) / y_pixels)
        h_data = np.maximum(data[data.hcol].to_numpy(), 0.5 * pix_size)
    else:
        h_data = data[data.hcol].to_numpy()

    return h_data


def interpolate_2d(data: 'SarracenDataFrame',  # noqa: F821
                   target: str,
                   x: Union[str, None] = None,
                   y: Union[str, None] = None,
                   kernel: Union[BaseKernel, None] = None,
                   x_pixels: Union[int, None] = None,
                   y_pixels: Union[int, None] = None,
                   xlim: Optional[Tuple[Optional[float],
                                        Optional[float]]] = None,
                   ylim: Optional[Tuple[Optional[float],
                                        Optional[float]]] = None,
                   exact: bool = False,
                   backend: Union[str, None] = None,
                   dens_weight: bool = False,
                   normalize: bool = True,
                   hmin: bool = False) -> np.ndarray:
    """
    Interpolate particle data across two directional axes to a 2D grid of
    pixels.

    Interpolate the data within a SarracenDataFrame to a 2D grid, by
    interpolating the values of a target variable. The contributions of all
    particles near the interpolation area are summed and stored to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target smoothing data.
    x, y: str, optional
        Column labels of the directional axes. Defaults to the x & y columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    exact: bool, optional
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to False (this may
        change in future versions).
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (2-Dimensional)
        The interpolated output image, in a 2-dimensional numpy array.
        Dimensions are structured in reverse order, where (x, y) -> [y, x].

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximum values result in an invalid
        region, or if `data` is not 2-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do
        not exist in `data`.
    """
    _check_dimension(data, 2)
    x, y = _default_xy(data, x, y)
    _verify_columns(data, x, y)

    xlim, ylim = _default_bounds(data[x], data[y], xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)
    w_data = _get_weight(data, target, dens_weight)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend)\
        .interpolate_2d_render(data[x].to_numpy(), data[y].to_numpy(),
                               w_data, h_data, kernel.w, kernel.get_radius(),
                               x_pixels, y_pixels, xlim[0], xlim[1],
                               ylim[0], ylim[1], exact)

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend)\
            .interpolate_2d_render(data[x].to_numpy(), data[y].to_numpy(),
                                   w_norm, h_data, kernel.w,
                                   kernel.get_radius(), x_pixels, y_pixels,
                                   xlim[0], xlim[1], ylim[0], ylim[1], exact)
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_2d_vec(data: 'SarracenDataFrame',  # noqa: F821
                       target_x: str,
                       target_y: str,
                       x: Union[str, None] = None,
                       y: Union[str, None] = None,
                       kernel: Union[BaseKernel, None] = None,
                       x_pixels: Union[int, None] = None,
                       y_pixels: Union[int, None] = None,
                       xlim: Optional[Tuple[Optional[float],
                                            Optional[float]]] = None,
                       ylim: Optional[Tuple[Optional[float],
                                            Optional[float]]] = None,
                       exact: bool = False,
                       backend: Union[str, None] = None,
                       dens_weight: bool = False,
                       normalize: bool = True,
                       hmin: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """
    Interpolate vector particle data across two directional axes to a 2D grid
    of particles.

    Interpolate the data within a SarracenDataFrame to a 2D grid, by
    interpolating the values of a target vector. The contributions of all
    vectors near the interpolation area are summed and stored to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target_x, target_y: str
        Column labels of the target vector.
    x, y: str, optional
        Column labels of the directional axes. Defaults to the x & y columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    exact: bool, optional
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to False (this may
        change in future versions).
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output_x, output_y: ndarray (2-Dimensional)
        The interpolated output images, in a 2-dimensional numpy arrays.
        Dimensions are structured in reverse order, where (x, y) -> [y, x].

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximum values result in an invalid
        region, or if `data` is not 2-dimensional.
    KeyError
        If `target_x`, `target_y`, `x`, `y`, mass, density, or smoothing
        length columns do not exist in `data`.
    """
    _check_dimension(data, 2)
    x, y = _default_xy(data, x, y)
    _verify_columns(data, x, y)

    xlim, ylim = _default_bounds(data[x], data[y], xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    wx_data = _get_weight(data, target_x, dens_weight)
    wy_data = _get_weight(data, target_y, dens_weight)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    gridx, gridy = get_backend(backend)\
        .interpolate_2d_render_vec(data[x].to_numpy(), data[y].to_numpy(),
                                   wx_data, wy_data, h_data, kernel.w,
                                   kernel.get_radius(), x_pixels, y_pixels,
                                   xlim[0], xlim[1], ylim[0], ylim[1], exact)

    if normalize:
        wx_norm = _get_weight(data, np.array([1] * len(wx_data)), dens_weight)
        wy_norm = _get_weight(data, np.array([1] * len(wy_data)), dens_weight)
        norm_gridx, norm_gridy = get_backend(backend)\
            .interpolate_2d_render_vec(data[x].to_numpy(), data[y].to_numpy(),
                                       wx_norm, wy_norm, h_data, kernel.w,
                                       kernel.get_radius(), x_pixels, y_pixels,
                                       xlim[0], xlim[1], ylim[0], ylim[1],
                                       exact)
        gridx = np.nan_to_num(gridx / norm_gridx)
        gridy = np.nan_to_num(gridy / norm_gridy)

    return (gridx, gridy)


def interpolate_2d_line(data: 'SarracenDataFrame',  # noqa: F821
                        target: str,
                        x: Union[str, None] = None,
                        y: Union[str, None] = None,
                        kernel: Union[BaseKernel, None] = None,
                        pixels: Union[int, None] = None,
                        xlim: Union[Tuple[float, float], None] = None,
                        ylim: Union[Tuple[float, float], None] = None,
                        backend: Union[str, None] = None,
                        dens_weight: bool = False,
                        normalize: bool = True,
                        hmin: bool = False) -> np.ndarray:
    """
    Interpolate particle data across two directional axes to a 1D cross-section
    line.

    Interpolate the data within a SarracenDataFrame to a 1D line, by
    interpolating the values of a target variable. The contributions of all
    particles near the specified line are summed and stored to a 1-dimensional
    array.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
         Column label of the target smoothing data.
    x, y: str, optional
        Column labels of the directional axes. Defaults to the x & y columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    pixels: int, optional
        Number of points in the resulting line plot in the x-direction.
    xlim, ylim: tuple of float, optional
        Starting and ending coordinates of the cross-section line (in particle
        data space). Defaults to the minimum and maximum values of `x` and `y`.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
         cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    np.ndarray (1-Dimensional)
        The resulting interpolated output.

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `xlim` and `ylim` values are all the same (indicating a
        zero-length cross-section), or if `data` is not 2-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do
        not exist in `data`.
    """
    _check_dimension(data, 2)
    x, y = _default_xy(data, x, y)
    _verify_columns(data, x, y)

    w_data = _get_weight(data, target, dens_weight)

    if isinstance(xlim, float) or isinstance(xlim, int):
        xlim = xlim, xlim
    if isinstance(ylim, float) or isinstance(ylim, int):
        ylim = ylim, ylim

    xlim, ylim = _default_bounds(data[x], data[y], xlim, ylim)
    if xlim[0] == xlim[1] and ylim[0] == ylim[1]:
        raise ValueError('Zero length cross section!')

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend
    pixels = pixels if pixels is not None else 512

    if pixels <= 0:
        raise ValueError('pixcount must be greater than zero!')

    if hmin:
        pix_size = np.sqrt((xlim[1] - xlim[0])**2
                           + (ylim[1] - ylim[0])**2) / pixels
        h_data = np.maximum(data[data.hcol].to_numpy(), 0.5 * pix_size)
    else:
        h_data = data[data.hcol].to_numpy()

    grid = get_backend(backend) \
        .interpolate_2d_line(data[x].to_numpy(), data[y].to_numpy(),
                             w_data, h_data, kernel.w, kernel.get_radius(),
                             pixels, xlim[0], xlim[1], ylim[0], ylim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
            .interpolate_2d_line(data[x].to_numpy(), data[y].to_numpy(),
                                 w_norm, h_data, kernel.w,
                                 kernel.get_radius(), pixels, xlim[0],
                                 xlim[1], ylim[0], ylim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_line(data: 'SarracenDataFrame',  # noqa: F821
                        target: str,
                        x: Union[str, None] = None,
                        y: Union[str, None] = None,
                        z: Union[str, None] = None,
                        kernel: Union[BaseKernel, None] = None,
                        pixels: Union[int, None] = None,
                        xlim: Union[Tuple[float, float], None] = None,
                        ylim: Union[Tuple[float, float], None] = None,
                        zlim: Union[Tuple[float, float], None] = None,
                        backend: Union[str, None] = None,
                        dens_weight: bool = False,
                        normalize: bool = True,
                        hmin: bool = False) -> np.ndarray:
    """
    Interpolate vector particle data across three directional axes to a 1D
    line.

    Interpolate the data within a SarracenDataFrame to a 1D line, by
    interpolating the values of a target variable. The contributions of all
    particles near the interpolation line are summed and stored to a 1D array.

    Parameters
    ----------
    data : SarracenDataFrame
            Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target variable.
    x, y, z: str, optional
        Column labels of the directional axes. Defaults to the x, y & z columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim, zlim: tuple of float, optional
        Starting and ending coordinates of the cross-section line (in particle
        data space). Defaults to the minimum and maximum values of `x`, `y`,
        and `z`.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
       If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output: ndarray (1-Dimensional)
        The interpolated output line.

    Raises
    ------
    ValueError
        If `pixels` are less than or equal to zero, or if the specified `x`,
        `y`, and `z` minimum and maximum values result in a zero area
        cross-section, or if `data` is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length data does
        not exist in `data`.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    w_data = _get_weight(data, target, dens_weight)

    if isinstance(xlim, float) or isinstance(xlim, int):
        xlim = xlim, xlim
    if isinstance(ylim, float) or isinstance(ylim, int):
        ylim = ylim, ylim
    if isinstance(zlim, float) or isinstance(zlim, int):
        zlim = zlim, zlim

    z1 = data.loc[:, z].min() if zlim is None or zlim[0] is None else zlim[0]
    z2 = data.loc[:, z].min() if zlim is None or zlim[1] is None else zlim[1]
    zlim = z1, z2

    xlim, ylim = _default_bounds(data[x], data[y], xlim, ylim)
    if ylim[1] == ylim[0] and xlim[1] == xlim[0] and zlim[1] == zlim[0]:
        raise ValueError('Zero length cross section!')

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend
    pixels = pixels if pixels is not None else 512

    if pixels <= 0:
        raise ValueError('pixcount must be greater than zero!')

    if hmin:
        pix_size = np.sqrt((xlim[1] - xlim[0])**2
                           + (ylim[1] - ylim[0])**2
                           + (zlim[1] - zlim[0])**2) / pixels
        h_data = np.maximum(data[data.hcol].to_numpy(), 0.5 * pix_size)
    else:
        h_data = data[data.hcol].to_numpy()

    grid = get_backend(backend) \
        .interpolate_3d_line(data[x].to_numpy(), data[y].to_numpy(),
                             data[z].to_numpy(), w_data, h_data, kernel.w,
                             kernel.get_radius(), pixels, xlim[0], xlim[1],
                             ylim[0], ylim[1], zlim[0], zlim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
            .interpolate_3d_line(data[x].to_numpy(), data[y].to_numpy(),
                                 data[z].to_numpy(), w_norm, h_data, kernel.w,
                                 kernel.get_radius(), pixels, xlim[0], xlim[1],
                                 ylim[0], ylim[1], zlim[0], zlim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_proj(data: 'SarracenDataFrame',  # noqa: F821
                        target: str,
                        x: Union[str, None] = None,
                        y: Union[str, None] = None,
                        kernel: Union[BaseKernel, None] = None,
                        integral_samples: int = 1000,
                        corotation: Union[np.ndarray, list, None] = None,
                        rotation: Union[np.ndarray, list,
                                        Rotation, None] = None,
                        rot_origin: Union[np.ndarray, list, str, None] = None,
                        x_pixels: Union[int, None] = None,
                        y_pixels: Union[int, None] = None,
                        xlim: Union[Tuple[float, float], None] = None,
                        ylim: Union[Tuple[float, float], None] = None,
                        exact: bool = False,
                        backend: Union[str, None] = None,
                        dens_weight: Union[bool, None] = None,
                        normalize: bool = True,
                        hmin: bool = False) -> np.ndarray:
    """
    Interpolate 3D particle data to a 2D grid of pixels.

    Interpolates three-dimensional particle data in a SarracenDataFrame. The
    data is interpolated to a 2D grid of pixels, by summing contributions in
    columns which span the z-axis.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target smoothing data.
    x, y: str, optional
        Column labels of the directional axes. Defaults to the x & y columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    integral_samples: int, optional
        Number of sample points to take when approximating the 2D column
        kernel.
    corotation: array_like, optional
        The x, y, z coordinates of two locations which determines the
        corotating frame.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    exact: bool, optional
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to True for
        column-integrated views, when the target is not density, and False for
        everything else.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (2-Dimensional)
        The interpolated output image, in a 2-dimensional numpy array.
        Dimensions are structured in reverse order, where (x, y) -> [y, x].

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximums result in an invalid region,
        or if the provided data is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do
        not exist in `data`.

    Notes
    -----
    Since the direction of integration is assumed to be straight across the
    z-axis, the z-axis column is not required for this type of interpolation.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, None)
    _verify_columns(data, x, y)

    if dens_weight is None:
        dens_weight = (target != 'rho')

    w_data = _get_weight(data, target, dens_weight)

    if corotation is not None:
        rotation, rot_origin = _corotate(corotation, rotation)

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, rot_origin)

    xlim, ylim = _default_bounds(x_data, y_data, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    weight_function = kernel.get_column_kernel_func(integral_samples)

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend) \
        .interpolate_3d_projection(x_data, y_data, w_data, h_data,
                                   weight_function, kernel.get_radius(),
                                   x_pixels, y_pixels,
                                   xlim[0], xlim[1], ylim[0], ylim[1], exact)

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
            .interpolate_3d_projection(x_data, y_data, w_norm, h_data,
                                       weight_function, kernel.get_radius(),
                                       x_pixels, y_pixels, xlim[0], xlim[1],
                                       ylim[0], ylim[1], exact)
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_vec(data: 'SarracenDataFrame',  # noqa: F821
                       target_x: str,
                       target_y: str,
                       target_z: str,
                       x: Union[str, None] = None,
                       y: Union[str, None] = None,
                       kernel: Union[BaseKernel, None] = None,
                       integral_samples: int = 1000,
                       rotation: Union[np.ndarray, list,
                                       Rotation, None] = None,
                       rot_origin: Union[np.ndarray, list, str, None] = None,
                       x_pixels: Union[int, None] = None,
                       y_pixels: Union[int, None] = None,
                       xlim: Optional[Tuple[Optional[float],
                                            Optional[float]]] = None,
                       ylim: Optional[Tuple[Optional[float],
                                            Optional[float]]] = None,
                       exact: bool = False,
                       backend: Union[str, None] = None,
                       dens_weight: bool = False,
                       normalize: bool = True,
                       hmin: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """
    Interpolate 3D vector particle data to a 2D grid of pixels.

    Interpolates three-dimensional vector particle data in a SarracenDataFrame.
    The data is interpolated to a 2D grid of pixels, by summing contributions
    in columns which span the z-axis.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target_x, target_y, target_z: str
        Column labels of the target vector.
    x, y: str, optional
        Column labels of the directional axes. Defaults to the x & y columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    integral_samples: int, optional
        Number of sample points to take when approximating the 2D column
        kernel.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    exact: bool, optional
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output_x, output_y: ndarray (2-Dimensional)
        The interpolated output images. Dimensions are structured in reverse
        order, where (x, y) -> [y, x].

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximums result in an invalid region,
        or if the provided data is not 3-dimensional.
    KeyError
        If `target_x`, `target_y`, `x`, `y`, mass, density, or smoothing
        length columns do not exist in `data`.

    Notes
    -----
    Since the direction of integration is assumed to be straight across the
    z-axis, the z-axis column is not required for this type of interpolation.
    """

    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, None)
    _verify_columns(data, x, y)

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, rot_origin)

    xlim, ylim = _default_bounds(x_data, y_data, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    if target_z not in data.columns:
        raise KeyError(f"z-directional target column '{target_z}' does not "
                       f"exist in the provided dataset.")
    target_x_data, target_y_data, _ = _rotate_data(data, data[target_x],
                                                   data[target_y],
                                                   data[target_z], rotation,
                                                   rot_origin)

    wx_data = _get_weight(data, target_x_data, dens_weight)
    wy_data = _get_weight(data, target_y_data, dens_weight)
    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    weight_function = kernel.get_column_kernel_func(integral_samples)
    gridx, gridy = get_backend(backend) \
        .interpolate_3d_projection_vec(x_data, y_data, wx_data, wy_data,
                                       h_data, weight_function,
                                       kernel.get_radius(), x_pixels, y_pixels,
                                       xlim[0], xlim[1], ylim[0], ylim[1],
                                       exact)
    if normalize:
        wx_norm = _get_weight(data, np.array([1] * len(wx_data)), dens_weight)
        wy_norm = _get_weight(data, np.array([1] * len(wy_data)), dens_weight)
        norm_gridx, norm_gridy = get_backend(backend) \
            .interpolate_3d_projection_vec(x_data, y_data, wx_norm, wy_norm,
                                           h_data, weight_function,
                                           kernel.get_radius(), x_pixels,
                                           y_pixels, xlim[0], xlim[1],
                                           ylim[0], ylim[1], exact)
        gridx = np.nan_to_num(gridx / norm_gridx)
        gridy = np.nan_to_num(gridy / norm_gridy)

    return (gridx, gridy)


def interpolate_3d_cross(data: 'SarracenDataFrame',  # noqa: F821
                         target: str,
                         x: Union[str, None] = None,
                         y: Union[str, None] = None,
                         z: Union[str, None] = None,
                         z_slice: Union[float, None] = None,
                         kernel: Union[BaseKernel, None] = None,
                         corotation: Union[np.ndarray, list, None] = None,
                         rotation: Union[np.ndarray, list,
                                         Rotation, None] = None,
                         rot_origin: Union[np.ndarray, list, str, None] = None,
                         x_pixels: Union[int, None] = None,
                         y_pixels: Union[int, None] = None,
                         xlim: Union[Tuple[float, float], None] = None,
                         ylim: Union[Tuple[float, float], None] = None,
                         backend: Union[str, None] = None,
                         dens_weight: bool = False,
                         normalize: bool = True,
                         hmin: bool = False) -> np.ndarray:
    """
    Interpolate 3D particle data to a 2D grid, using a 3D cross-section.

    Interpolates particle data in a SarracenDataFrame across three directional
    axes to a 2D grid of pixels. A cross-section is taken of the 3D data at a
    specific value of z, and the contributions of particles near the plane are
    interpolated to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        The particle data to interpolate over.
    target: str
        The column label of the target smoothing data.
    x, y, z: str, optional
        The column labels of the directional data to interpolate over. Defaults
        to the x, y, and z columns
        detected in `data`.
    z_slice: float, optional
        The z-axis value to take the cross-section at. Defaults to the midpoint
        of the z-directional data.
    kernel: BaseKernel, optional
        The kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    corotation: array_like, optional
        The x, y, z coordinates of two locations which determines the
        corotating frame.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (2-Dimensional)
        The interpolated output image, in a 2-dimensional numpy array.
        Dimensions are structured in reverse order, where (x, y) -> [y, x].

    Raises
    ------
    ValueError
        If `pixwidthx`, `pixwidthy`, `pixcountx`, or `pixcounty` are less than
        or equal to zero, or if the specified `x` and `y` minimum and maximums
        result in an invalid region, or if the provided data is not
        3-dimensional.
    KeyError
        If `target`, `x`, `y`, `z`, mass, density, or smoothing length columns
        do not exist in `data`.
    """
    _check_dimension(data, 3)

    # x and y columns default to the variables from the SarracenDataFrame.
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    # set default slice to be through the data's average z-value.
    if z_slice is None:
        z_slice = data.loc[:, z].mean()

    w_data = _get_weight(data, target, dens_weight)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    if corotation is not None:
        rotation, rot_origin = _corotate(corotation, rotation)

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, rot_origin)

    # boundaries of the plot default to the max & min values of the data.
    xlim, ylim = _default_bounds(x_data, y_data, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend) \
        .interpolate_3d_cross(x_data, y_data, z_data, z_slice, w_data, h_data,
                              kernel.w, kernel.get_radius(), x_pixels,
                              y_pixels, xlim[0], xlim[1], ylim[0], ylim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
            .interpolate_3d_cross(x_data, y_data, z_data, z_slice, w_norm,
                                  h_data, kernel.w, kernel.get_radius(),
                                  x_pixels, y_pixels,
                                  xlim[0], xlim[1], ylim[0], ylim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_cross_vec(data: 'SarracenDataFrame',  # noqa: F821
                             target_x: str,
                             target_y: str,
                             target_z: str,
                             z_slice: Union[float, None] = None,
                             x: Union[str, None] = None,
                             y: Union[str, None] = None,
                             z: Union[str, None] = None,
                             kernel: Union[BaseKernel, None] = None,
                             rotation: Union[np.ndarray, list, Rotation,
                                             None] = None,
                             rot_origin: Union[np.ndarray, list,
                                               str, None] = None,
                             x_pixels: Union[int, None] = None,
                             y_pixels: Union[int, None] = None,
                             xlim: Optional[Tuple[Optional[float],
                                                  Optional[float]]] = None,
                             ylim: Optional[Tuple[Optional[float],
                                                  Optional[float]]] = None,
                             backend: Union[str, None] = None,
                             dens_weight: bool = False,
                             normalize: bool = True,
                             hmin: bool = False) -> Tuple[np.ndarray,
                                                          np.ndarray]:
    """
    Interpolate 3D vector particle data to a 2D grid, using a 3D cross-section.

    Interpolates vector particle data in a SarracenDataFrame across three
    directional axes to a 2D grid of pixels. A cross-section is taken of the
    3D data at a specific value of z, and the contributions of vectors near
    the plane are interpolated to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        The particle data to interpolate over.
    target_x, target_y, target_z: str
        The column labels of the target vector.
    z_slice: float, optional
        The z-axis value to take the cross-section at. Defaults to the midpoint
        of the z-directional data.
    x, y, z: str, optional
        The column labels of the directional data to interpolate over. Defaults
        to the x, y, and z columns detected in `data`.
    kernel: BaseKernel, optional
        The kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output_x, output_y: ndarray (2-Dimensional)
        The interpolated output images. Dimensions are structured in reverse
        order, where (x, y) -> [y, x].

    Raises
    ------
    ValueError
        If `pixwidthx`, `pixwidthy`, `pixcountx`, or `pixcounty` are less than
        or equal to zero, or if the specified `x` and `y` minimum and maximums
        result in an invalid region, or if the provided data is not
        3-dimensional.
    KeyError
        If `target_x`, `target_y`, `target_z`, `x`, `y`, `z`, mass, density,
        or smoothing length columns do not exist in `data`.
    """

    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    # set default slice to be through the data's average z-value.
    if z_slice is None:
        z_slice = data.loc[:, z].mean()

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, rot_origin)

    # boundaries of the plot default to the max & min values of the data.
    xlim, ylim = _default_bounds(x_data, y_data, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    target_x_data, target_y_data, _ = _rotate_data(data, data[target_x],
                                                   data[target_y],
                                                   data[target_z], rotation,
                                                   rot_origin)

    wx_data = _get_weight(data, target_x_data, dens_weight)
    wy_data = _get_weight(data, target_y_data, dens_weight)
    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    gridx, gridy = get_backend(backend) \
        .interpolate_3d_cross_vec(x_data, y_data, z_data, z_slice, wx_data,
                                  wy_data, h_data, kernel.w,
                                  kernel.get_radius(), x_pixels, y_pixels,
                                  xlim[0], xlim[1], ylim[0], ylim[1])

    if normalize:
        wx_norm = _get_weight(data, np.array([1] * len(wx_data)), dens_weight)
        wy_norm = _get_weight(data, np.array([1] * len(wy_data)), dens_weight)
        norm_gridx, norm_gridy = get_backend(backend) \
            .interpolate_3d_cross_vec(x_data, y_data, z_data, z_slice, wx_norm,
                                      wy_norm, h_data, kernel.w,
                                      kernel.get_radius(), x_pixels, y_pixels,
                                      xlim[0], xlim[1], ylim[0], ylim[1])
        gridx = np.nan_to_num(gridx / norm_gridx)
        gridy = np.nan_to_num(gridy / norm_gridy)

    return (gridx, gridy)


def interpolate_3d_grid(data: 'SarracenDataFrame',  # noqa: F821
                        target: str,
                        x: Union[str, None] = None,
                        y: Union[str, None] = None,
                        z: Union[str, None] = None,
                        kernel: Union[BaseKernel, None] = None,
                        rotation: Union[np.ndarray, list,
                                        Rotation, None] = None,
                        rot_origin: Union[np.ndarray, list, str, None] = None,
                        x_pixels: Union[int, None] = None,
                        y_pixels: Union[int, None] = None,
                        z_pixels: Union[int, None] = None,
                        xlim: Optional[Tuple[Optional[float],
                                             Optional[float]]] = None,
                        ylim: Optional[Tuple[Optional[float],
                                             Optional[float]]] = None,
                        zlim: Union[Tuple[float, float], None] = None,
                        backend: Union[str, None] = None,
                        dens_weight: bool = False,
                        normalize: bool = True,
                        hmin: bool = False) -> np.ndarray:
    """
    Interpolate 3D particle data to a 3D grid of pixels

    Interpolates particle data in a SarracenDataFrame across three directional
    axes to a 3D grid of pixels. The contributions of all particles near each
    3D cell are summed and stored in the 3D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        The particle data to interpolate over.
    target: str
        The column label of the target data.
    x, y, z: str, optional
        The column labels of the directional data to interpolate over. Defaults
        to the x, y, and z columns detected in `data`.
    kernel: BaseKernel, optional
        The kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_pixels, y_pixels, z_pixels: int, optional
        Number of pixels in the output image in the x, y & z directions.
        Default values are chosen to keep a consistent aspect ratio.
    xlim, ylim, zlim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x`, `y`
        and `z`.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, the target will be multiplied by density. Defaults to False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (3-Dimensional)
        The interpolated output image, in a 3-dimensional numpy array.
        Dimensions are structured in reverse order, where (x, y, z) ->
        [z, y, x].

    Raises
    ------
    ValueError
        If `x_pixels`, `y_pixels` or `z_pixels` are less than or equal to zero,
        or if the specified `x`, `y` and `z` minimum and maximum values result
        in an invalid region, or if `data` is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, `z`, mass, density, or smoothing length columns
        do not exist in `data`.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    w_data = _get_weight(data, target, dens_weight)

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, rot_origin)
    if not xlim:
        xlim = (None, None)
    if not ylim:
        ylim = (None, None)
    xlim, ylim = _default_bounds(x_data, y_data, xlim, ylim)
    zlim = zlim if zlim else (z_data.min(), z_data.max())

    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    if z_pixels is None:
        dz = zlim[1] - zlim[0]
        dx = xlim[1] - xlim[0]
        z_pixels = int(np.rint(x_pixels * (dz / dx)))

    _check_boundaries(x_pixels, y_pixels, xlim, ylim)
    if zlim[1] - zlim[0] <= 0:
        raise ValueError("`z_max` must be greater than `z_min`!")
    if z_pixels <= 0:
        raise ValueError("`z_pixels` must be greater than zero!")

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels,
                                    xlim, ylim)

    grid = get_backend(backend)\
        .interpolate_3d_grid(x_data, y_data, z_data, w_data, h_data, kernel.w,
                             kernel.get_radius(), x_pixels, y_pixels, z_pixels,
                             xlim[0], xlim[1], ylim[0], ylim[1],
                             zlim[0], zlim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend)\
            .interpolate_3d_grid(x_data, y_data, z_data, w_norm, h_data,
                                 kernel.w, kernel.get_radius(), x_pixels,
                                 y_pixels, z_pixels, xlim[0], xlim[1], ylim[0],
                                 ylim[1], zlim[0], zlim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def get_backend(code: str) -> Type[BaseBackend]:
    """
    Get the interpolation backend associated with a string code.

    Parameters
    ----------
    code: str
        The code associated with the particular backend. At the moment, 'cpu'
        for the CPU backend, and 'gpu' for the GPU backend are supported.

    Returns
    -------
    BaseBackend: The backend to use for interpolation.
    """
    if code == 'cpu':
        return CPUBackend
    if code == 'gpu':
        return GPUBackend
    raise ValueError("Invalid backend")



from typing import Tuple

from numba.core.registry import CPUDispatcher
from numpy import ndarray, zeros


class BaseBackend:
    """Backend implementation of SPH interpolation functions."""

    @staticmethod
    def interpolate_2d_render(x: ndarray,
                              y: ndarray,
                              weight: ndarray,
                              h: ndarray,
                              weight_function: CPUDispatcher,
                              kernel_radius: float,
                              x_pixels: int,
                              y_pixels: int,
                              x_min: float,
                              x_max: float,
                              y_min: float,
                              y_max: float,
                              exact: bool) -> ndarray:
        """ Interpolate 2D data to a 2D grid of pixels."""
        return zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_2d_render_vec(x: ndarray,
                                  y: ndarray,
                                  weight_x: ndarray,
                                  weight_y: ndarray,
                                  h: ndarray,
                                  weight_function: CPUDispatcher,
                                  kernel_radius: float,
                                  x_pixels: int,
                                  y_pixels: int,
                                  x_min: float,
                                  x_max: float,
                                  y_min: float,
                                  y_max: float,
                                  exact: bool) -> Tuple[ndarray, ndarray]:
        """ Interpolate 2D vector data to a pair of 2D pixel grids. """
        return zeros((y_pixels, x_pixels)), zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_2d_line(x: ndarray,
                            y: ndarray,
                            weight: ndarray,
                            h: ndarray,
                            weight_function: CPUDispatcher,
                            kernel_radius: float,
                            pixels: int,
                            x1: float,
                            x2: float,
                            y1: float,
                            y2: float) -> ndarray:
        """ Interpolate 2D data to a 1D cross-sectional line. """
        return zeros(pixels)

    @staticmethod
    def interpolate_3d_line(x: ndarray,
                            y: ndarray,
                            z: ndarray,
                            weight: ndarray,
                            h: ndarray,
                            weight_function: CPUDispatcher,
                            kernel_radius: float,
                            pixels: int,
                            x1: float,
                            x2: float,
                            y1: float,
                            y2: float,
                            z1: float,
                            z2: float) -> ndarray:
        """ Interpolate 3D data to a 1D cross-sectional line. """
        return zeros(pixels)

    @staticmethod
    def interpolate_3d_projection(x: ndarray,
                                  y: ndarray,
                                  weight: ndarray,
                                  h: ndarray,
                                  weight_function: CPUDispatcher,
                                  kernel_radius: float,
                                  x_pixels: int,
                                  y_pixels: int,
                                  x_min: float,
                                  x_max: float,
                                  y_min: float,
                                  y_max: float,
                                  exact: bool) -> ndarray:
        """ Interpolate 3D data to a 2D pixel grid using column projection."""
        return zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_projection_vec(x: ndarray,
                                      y: ndarray,
                                      weight_x: ndarray,
                                      weight_y: ndarray,
                                      h: ndarray,
                                      weight_function: CPUDispatcher,
                                      kernel_radius: float,
                                      x_pixels: int,
                                      y_pixels: int,
                                      x_min: float,
                                      x_max: float,
                                      y_min: float,
                                      y_max: float,
                                      exact: bool) -> Tuple[ndarray, ndarray]:
        """ Interpolate 3D vector data to a pair of 2D pixel grids using
        column projection."""
        return zeros((y_pixels, x_pixels)), zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_cross(x: ndarray,
                             y: ndarray,
                             z: ndarray,
                             z_slice: float,
                             weight: ndarray,
                             h: ndarray,
                             weight_function: CPUDispatcher,
                             kernel_radius: float,
                             x_pixels: int,
                             y_pixels: int,
                             x_min: float,
                             x_max: float,
                             y_min: float,
                             y_max: float) -> ndarray:
        """
        Interpolate 3D data to a pair of 2D pixel grids using a 3D
        cross-section at a specific z value.
        """
        return zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_cross_vec(x: ndarray,
                                 y: ndarray,
                                 z: ndarray,
                                 z_slice: float,
                                 weight_x: ndarray,
                                 weight_y: ndarray,
                                 h: ndarray,
                                 weight_function: CPUDispatcher,
                                 kernel_radius: float,
                                 x_pixels: int,
                                 y_pixels: int,
                                 x_min: float,
                                 x_max: float,
                                 y_min: float,
                                 y_max: float) -> Tuple[ndarray, ndarray]:
        """
        Interpolate 3D vector data to a pair of 2D pixel grids using a 3D
        cross-section at a specific z value.
        """
        return zeros((y_pixels, x_pixels)), zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_grid(x: ndarray,
                            y: ndarray,
                            z: ndarray,
                            weight: ndarray,
                            h: ndarray,
                            weight_function: CPUDispatcher,
                            kernel_radius: float,
                            x_pixels: int,
                            y_pixels: int,
                            z_pixels: int,
                            x_min: float,
                            x_max: float,
                            y_min: float,
                            y_max: float,
                            z_min: float,
                            z_max: float) -> ndarray:
        """
        Interpolate 3D data to a 3D grid of pixels.
        """
        return zeros((z_pixels, y_pixels, x_pixels))
